
import os
import unittest
import warnings
from pathlib import Path

import paramak
import pytest


class TestCenterColumnStudyReactor(unittest.TestCase):
    """Test functionality of the CenterColumnStudyReactor class"""

    def setUp(self):
        self.test_reactor = paramak.CenterColumnStudyReactor(
            inner_bore_radial_thickness=20,
            inboard_tf_leg_radial_thickness=50,
            center_column_shield_radial_thickness_mid=50,
            center_column_shield_radial_thickness_upper=100,
            inboard_firstwall_radial_thickness=20,
            divertor_radial_thickness=100,
            inner_plasma_gap_radial_thickness=80,
            plasma_radial_thickness=200,
            outer_plasma_gap_radial_thickness=90,
            elongation=2.3,
            triangularity=0.45,
            plasma_gap_vertical_thickness=40,
            center_column_arc_vertical_thickness=520,
            rotation_angle=359
        )

    def test_input_variable_names(self):
        """tests that the number of inputs variables is correct"""

        assert len(self.test_reactor.input_variables.keys()) == 17
        assert len(self.test_reactor.input_variable_names) == 17

    def test_creation(self):
        """Creates a ball reactor using the CenterColumnStudyReactor parametric_reactor and checks
        the correct number of components are created."""

        assert len(self.test_reactor.shapes_and_components) == 6

    def test_creation_with_narrow_divertor(self):
        """Creates a ball reactor with a narrow divertor using the CenterColumnStudyReactor
        parametric reactor and checks that the correct number of components are created."""

        self.test_reactor.divertor_radial_thickness = 10
        assert len(self.test_reactor.shapes_and_components) == 6

    def test_svg_creation(self):
        """Creates a ball reactor using the CenterColumnStudyReactor parametric_reactor and checks
        an svg image of the reactor can be exported."""

        os.system("rm test_image.svg")
        self.test_reactor.export_svg("test_image.svg")
        assert Path("test_image.svg").exists() is True
        os.system("rm test_image.svg")

    def test_rotation_angle_impacts_volume(self):
        """Creates a CenterColumnStudyReactor reactor with a rotation angle of
        90 and another reactor with a rotation angle of 180. Then checks the
        volumes of all the components is double in the 180 reactor."""

        self.test_reactor.rotation_angle = 90
        r90_comp_vols = [
            comp.volume() for comp in self.test_reactor.shapes_and_components]
        self.test_reactor.rotation_angle = 180
        r180_comp_vols = [
            comp.volume() for comp in self.test_reactor.shapes_and_components]
        for r90_vol, r180_vol in zip(r90_comp_vols, r180_comp_vols):
            assert r90_vol == pytest.approx(r180_vol * 0.5, rel=0.1)

    def test_rotation_angle_warning(self):
        """Checks that the correct warning message is printed when
        rotation_angle = 360."""

        def warning_trigger():
            try:
                self.test_reactor.rotation_angle = 360
                self.test_reactor.shapes_and_components
            except BaseException:
                pass

        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            warning_trigger()
            assert len(w) == 1
            assert issubclass(w[-1].category, UserWarning)
            assert "360 degree rotation may result in a Standard_ConstructionError or AttributeError" in str(
                w[-1].message)




import unittest

import paramak


class TestSegmentedBlanketBallReactor(unittest.TestCase):
    """Tests functionality of the SegmentedBlanketBallReactor class"""

    def setUp(self):
        self.test_reactor = paramak.SegmentedBlanketBallReactor(
            inner_bore_radial_thickness=10,
            inboard_tf_leg_radial_thickness=30,
            center_column_shield_radial_thickness=60,
            divertor_radial_thickness=150,
            inner_plasma_gap_radial_thickness=30,
            plasma_radial_thickness=300,
            outer_plasma_gap_radial_thickness=30,
            plasma_gap_vertical_thickness=30,
            firstwall_radial_thickness=20,
            blanket_radial_thickness=50,
            blanket_rear_wall_radial_thickness=30,
            elongation=2,
            triangularity=0.55,
            number_of_tf_coils=16,
            rotation_angle=180,
            pf_coil_radial_thicknesses=[50, 50, 50, 50],
            pf_coil_vertical_thicknesses=[50, 50, 50, 50],
            pf_coil_radial_position=[200, 200, 200, 200],
            pf_coil_vertical_position=[200, 100, -100, -200],
            rear_blanket_to_tf_gap=50,
            outboard_tf_coil_radial_thickness=100,
            outboard_tf_coil_poloidal_thickness=50,
            gap_between_blankets=30,
            number_of_blanket_segments=4,
        )

    def test_input_variable_names(self):
        """tests that the number of inputs variable is correct"""

        assert len(self.test_reactor.input_variables.keys()) == 31
        assert len(self.test_reactor.input_variable_names) == 31

    def test_gap_between_blankets_impacts_volume(self):
        """Creates a SegmentedBlanketBallReactor with different
        gap_between_blankets and checks the volume of the blankes and the
        firstwall changes."""

        self.test_reactor.create_solids()
        small_gap_blanket_volume = self.test_reactor._blanket.volume()
        small_gap_firstwall_volume = self.test_reactor._firstwall.volume()

        self.test_reactor.gap_between_blankets = 60
        self.test_reactor.create_solids()
        large_gap_blanket_volume = self.test_reactor._blanket.volume()
        large_gap_firstwall_volume = self.test_reactor._firstwall.volume()

        assert small_gap_blanket_volume > large_gap_blanket_volume
        assert small_gap_firstwall_volume > large_gap_firstwall_volume

    def test_number_of_blanket_segments_impacts_volume(self):
        """Creates a SegmentedBlanketBallReactor with different
        number_of_blanket_segments and checks the volume of the blanket and
        firstwall changes."""

        self.test_reactor.create_solids()
        blanket_volume_few_segments = self.test_reactor._blanket.volume()
        firstwall_volume_few_segments = self.test_reactor._firstwall.volume()

        self.test_reactor.number_of_blanket_segments = 6
        self.test_reactor.create_solids()
        blanket_volume_many_segments = self.test_reactor._blanket.volume()
        firstwall_volume_many_segments = self.test_reactor._firstwall.volume()

        assert blanket_volume_many_segments < blanket_volume_few_segments
        assert firstwall_volume_many_segments > firstwall_volume_few_segments

    def test_invalid_parameter_error_raises(self):
        """Checks that the correct errors are raised when invalid arguments for
        parameters are input."""

        def invalid_gap_between_blankets():
            self.test_reactor.gap_between_blankets = -1

        def invalid_number_of_blanket_segments():
            self.test_reactor.number_of_blanket_segments = 1

        self.assertRaises(ValueError, invalid_gap_between_blankets)
        self.assertRaises(ValueError, invalid_number_of_blanket_segments)




import warnings
from typing import Optional

from paramak import ExtrudeMixedShape, ExtrudeStraightShape
from paramak.utils import add_thickness, cut_solid, union_solid


class TFCoilCasing(ExtrudeMixedShape):
    """Casing component for TF coils

    Args:
        magnet: TF coil shape
        inner_offset: radial distance between inner coil surface and inner
            casing surface (cm)
        outer_offset: radial distance between outer coil surface and outer
            casing surface (cm)
        vertical_section_offset: radial distance between outer coil surface and
            outer vertical section surface (cm)
        distance: extrusion distance (cm)
    """

    def __init__(
        self,
        magnet,
        inner_offset: float,
        outer_offset: float,
        vertical_section_offset: float,
        **kwargs
    ) -> None:

        self.magnet = magnet

        super().__init__(
            **kwargs
        )

        self.inner_offset = inner_offset
        self.outer_offset = outer_offset
        self.vertical_section_offset = vertical_section_offset
        self.leg_shape = ExtrudeStraightShape(
            distance=self.distance,
            azimuth_placement_angle=self.azimuth_placement_angle,
            workplane=self.magnet.workplane)
        self.inner_bore_cutting_points = None
        self.workplane = self.magnet.workplane

    @property
    def azimuth_placement_angle(self):
        self.azimuth_placement_angle = self.magnet.azimuth_placement_angle
        return self._azimuth_placement_angle

    @azimuth_placement_angle.setter
    def azimuth_placement_angle(self, value):
        correct_angles = self.magnet.azimuth_placement_angle
        if value != correct_angles:
            msg = ('Casing azimuth_placement_angle should be the'
                   'same value as TFCoilCasing.magnet.')
            warnings.warn(msg, UserWarning)
        self._azimuth_placement_angle = correct_angles

    def find_points(self):
        inner_points_magnet = self.magnet.inner_points
        outer_points_magnet = self.magnet.outer_points
        inner_points_magnet = (
            inner_points_magnet[:, 0],
            inner_points_magnet[:, 1]
        )
        outer_points_magnet = (
            outer_points_magnet[:, 0],
            outer_points_magnet[:, 1]
        )
        inner_points = add_thickness(
            *inner_points_magnet,
            thickness=-self.inner_offset,
        )

        outer_points = add_thickness(
            *outer_points_magnet,
            thickness=-self.outer_offset
        )
        curve_points = []
        for distrib_points in [inner_points, outer_points]:
            curve_points.append([[R, Z, 'spline'] for R, Z in zip(
                distrib_points[0], distrib_points[1])])

        curve_points[0][-1][2] = 'straight'
        curve_points[1][-1][2] = "straight"

        points = curve_points[0] + curve_points[1]
        self.points = points

        yA = outer_points[1][outer_points[0].index(
            min(outer_points[0], key=lambda x:abs(x - min(inner_points[0]))))]
        self.leg_points = [
            (
                min(outer_points[0]) -
                self.vertical_section_offset + self.outer_offset,
                min(outer_points[1])),
            (
                outer_points[0][outer_points[1].index(min(outer_points[1]))],
                min(outer_points[1])),
            (
                inner_points[0][inner_points[1].index(min(inner_points[1]))],
                min(inner_points[1])),
            (min(inner_points[0]), min(
                yA, self.magnet.vertical_displacement - yA)),
            # not having this line avoid unexpected surfaces
            (inner_points[0][-1], inner_points[1][-1]),
            (min(inner_points[0]), max(
                yA, self.magnet.vertical_displacement - yA)),
            (
                inner_points[0][inner_points[1].index(min(inner_points[1]))],
                max(inner_points[1])),
            (
                outer_points[0][outer_points[1].index(min(outer_points[1]))],
                max(outer_points[1])),
            (
                min(outer_points[0]) - \
                self.vertical_section_offset + self.outer_offset,
                max(outer_points[1])),
        ]

        self.inner_bore_cutting_points = [
            (0, self.leg_points[0][1]),
            self.leg_points[0],
            self.leg_points[-1],
            (0, self.leg_points[-1][1])
        ]

    def create_solid(self):
        solid = super().create_solid()

        self.leg_shape.points = self.leg_points
        self.leg_shape.distance = self.distance
        self.leg_shape.azimuth_placement_angle = \
            self.azimuth_placement_angle
        solid = union_solid(solid, self.leg_shape)
        solid = cut_solid(solid, self.magnet)

        # cut away excess casing outlined in #789
        inner_bore_cutting_shape = ExtrudeStraightShape(
            points=self.inner_bore_cutting_points,
            distance=self.distance,
            azimuth_placement_angle=self.azimuth_placement_angle
        )
        solid = cut_solid(solid, inner_bore_cutting_shape)

        self.solid = solid
        return solid



from typing import Optional, Tuple
from paramak import RotateMixedShape
import math


class CapsuleVacuumVessel(RotateMixedShape):
    """A cylindrical vessel volume with constant thickness that has addition
    spherical edges.

    Arguments:
        outer_start_point: the x,z coordinates of the outer bottom of the
            vacuum vessel
        radius: the radius from which the centres of the vessel meets the outer
            circumference.
        thickness: the radial thickness of the vessel in cm.
    """

    def __init__(
        self,
        radius: float,
        outer_start_point: Tuple[float, float],
        thickness: float,
        **kwargs
    ):
        self.radius = radius
        self.thickness = thickness
        self.outer_start_point = outer_start_point[0], outer_start_point[1]

        super().__init__(**kwargs)

    @property
    def radius(self):
        return self._radius

    @radius.setter
    def radius(self, value):
        if not isinstance(value, (float, int)):
            raise ValueError(
                'VacuumVessel.radius must be a number. Not', value)
        if value <= 0:
            msg = ('VacuumVessel.radius must be a positive number above 0. '
                   f'Not {value}')
            raise ValueError(msg)
        self._radius = value

    @property
    def thickness(self):
        return self._thickness

    @thickness.setter
    def thickness(self, value):
        if not isinstance(value, (float, int)):
            msg = f'VacuumVessel.thickness must be a number. Not {value}'
            raise ValueError(msg)
        if value <= 0:
            msg = (
                f'VacuumVessel.thickness must be a positive number above 0. Not {value}')
            raise ValueError(msg)
        self._thickness = value

    def find_points(self):
        """
        Finds the XZ points joined by straight and circle connections that describe the
        2D profile of the vessel shape.
        """

        radius = self.radius
        thickness = self.thickness
        bottom_outer_x, bottom_outer_y = self.outer_start_point
        top_outer_y = bottom_outer_y + (4 * radius)
        top_outer_x = bottom_outer_x
        inner_r = radius - thickness
        bottom_outer_x, bottom_outer_y, thickness, radius, top_outer_x, top_outer_y, inner_r = float(bottom_outer_x), float(
            bottom_outer_y), float(thickness), float(radius), float(top_outer_x), float(top_outer_y), float(inner_r)

        p1 = (bottom_outer_x, bottom_outer_y, 'circle')
        p3 = (p1[0] + radius, p1[1] + radius, 'straight')
        p4 = (p3[0], p3[1] + radius * 2, 'circle')
        p6 = (top_outer_x, top_outer_y, 'straight')
        p7 = (p6[0], p6[1] - thickness, 'circle')
        p9 = (p4[0] - thickness, p4[1], 'straight')
        p10 = (p3[0] - thickness, p3[1], 'circle')
        p12 = (p1[0], p1[1] + thickness, 'straight')
        p2 = ((p1[0]) + (radius * math.cos((3 * math.pi) / 8)), (p1[1] +
                                                                 radius) - (radius * math.sin((3 * math.pi) / 8)), 'circle')
        p5 = ((p6[0] + (radius * math.cos((2 * math.pi) / 8))), (p6[1] -
                                                                 radius) + (radius * math.sin((2 * math.pi) / 8)), 'circle')
        p8 = ((p7[0] + (inner_r * math.cos((2 * math.pi) / 8))), (p7[1] -
                                                                  inner_r) + (inner_r * math.sin((2 * math.pi) / 8)), 'circle')
        p11 = ((p12[0]) + (inner_r * math.cos((3 * math.pi) / 8)),
               (p12[1] + inner_r) - (inner_r * math.sin((3 * math.pi) / 8)),
               'circle')

        self.points = [
            p1,
            p2,
            p3,
            p4,
            p5,
            p6,
            p7,
            p8,
            p9,
            p10,
            p11,
            p12
        ]




from typing import Optional, Tuple
from paramak import RotateMixedShape


class BlanketConstantThicknessArcH(RotateMixedShape):
    """An outboard blanket volume that follows the curvature of a circular
    arc and a constant blanket thickness. The upper and lower edges continue
    horizontally for the thickness of the blanket to back of the blanket.

    Arguments:
        inner_mid_point: the x,z coordinates of the mid point on the inner
            surface of the blanket.
        inner_upper_point: the x,z coordinates of the upper point on the
            inner surface of the blanket.
        inner_lower_point: the x,z coordinates of the lower point on the
            inner surface of the blanket.
        thickness: the radial thickness of the blanket in cm.
    """

    def __init__(
        self,
        inner_mid_point: Tuple[float, float],
        inner_upper_point: Tuple[float, float],
        inner_lower_point: Tuple[float, float],
        thickness: float,
        **kwargs
    ) -> None:

        super().__init__(**kwargs)

        self.inner_upper_point = inner_upper_point
        self.inner_lower_point = inner_lower_point
        self.inner_mid_point = inner_mid_point
        self.thickness = thickness

    def find_points(self):

        self.points = [
            (self.inner_upper_point[0], self.inner_upper_point[1], "circle"),
            (self.inner_mid_point[0], self.inner_mid_point[1], "circle"),
            (self.inner_lower_point[0], self.inner_lower_point[1], "straight"),
            (
                self.inner_lower_point[0] + abs(self.thickness),
                self.inner_lower_point[1],
                "circle",
            ),
            (
                self.inner_mid_point[0] + abs(self.thickness),
                self.inner_mid_point[1],
                "circle",
            ),
            (
                self.inner_upper_point[0] + abs(self.thickness),
                self.inner_upper_point[1],
                "straight",
            )
        ]




import unittest

import paramak


class TestShellFS(unittest.TestCase):

    def setUp(self):
        # this is the shape that gets shelled
        self.test_shape = paramak.PoloidalFieldCoil(
            height=50,
            width=50,
            rotation_angle=90,
            center_point=(100, 100)
        )

    def test_creation_with_different_kinds(self):

        casing_int = paramak.ShellFS(
            shape=self.test_shape, kind='intersection')
        casing_int.export_stp('int.stp')
        casing_arc = paramak.ShellFS(shape=self.test_shape, kind='arc')
        casing_arc.export_stp('arc.stp')

        assert casing_int.solid is not None
        assert casing_int.solid is not None
        assert casing_int.volume() > casing_arc.volume()

    def test_creation_with_different_thickness(self):

        casing_small = paramak.ShellFS(shape=self.test_shape, thickness=5)

        casing_large = paramak.ShellFS(shape=self.test_shape, thickness=20)

        assert casing_small.solid is not None
        assert casing_large.solid is not None
        assert casing_large.volume() > casing_small.volume()




import unittest

import paramak
import pytest


class TestExtrudeRectangle(unittest.TestCase):

    def setUp(self):
        self.test_shape = paramak.ExtrudeRectangle(
            height=50, width=60, center_point=(1000, 500), distance=333
        )

    def test_default_parameters(self):
        """Checks that the default parameters of a ExtrudeRectangle are
        correct."""

        assert self.test_shape.distance == 333

    def test_processed_points_calculation(self):
        """Checks that the processed_points used to construct the
        ExtrudeRectangle are calculated correctly from the parameters given."""

        assert self.test_shape.processed_points == [
            (1030.0, 525.0, 'straight'),
            (1030.0, 475.0, 'straight'),
            (970.0, 475.0, 'straight'),
            (970.0, 525.0, 'straight'),
            (1030.0, 525.0, 'straight')
        ]

    def test_points_calculation(self):
        """Checks that the points used to construct the ExtrudeRectangle are
        calculated correctly from the parameters given."""

        assert self.test_shape.points == [
            (1030.0, 525.0),
            (1030.0, 475.0),
            (970.0, 475.0),
            (970.0, 525.0)
        ]

    def test_creation(self):
        """Creates a rectangular extrusion using the ExtrudeRectangle
        parametric component and checks that a cadquery solid is created."""

        assert self.test_shape.solid is not None
        assert self.test_shape.volume() > 1000

    def test_absolute_volume(self):
        """Creates a rectangular extrusion using the ExtrudeRectangle
        parametric component and checks that the volume is correct"""

        assert self.test_shape.volume() == pytest.approx(50 * 60 * 333)

    def test_absolute_areas(self):
        """Creates a rectangular extrusion using the ExtrudeRectangle
        parametric component and checks that the areas are correct"""

        assert len(self.test_shape.areas) == 6
        assert len(set([round(i) for i in self.test_shape.areas])) == 3
        assert self.test_shape.areas.count(
            pytest.approx(60 * 50)) == 2
        assert self.test_shape.areas.count(
            pytest.approx(50 * 333)) == 2
        assert self.test_shape.areas.count(
            pytest.approx(60 * 333)) == 2




import unittest

import paramak
import pytest


class TestSingleNullSubmersionTokamak(unittest.TestCase):
    """Tests functionality of SingleNullSubmersionTokamak class"""

    def setUp(self):
        self.test_reactor = paramak.SingleNullSubmersionTokamak(
            inner_bore_radial_thickness=30,
            inboard_tf_leg_radial_thickness=30,
            center_column_shield_radial_thickness=30,
            divertor_radial_thickness=80,
            inner_plasma_gap_radial_thickness=50,
            plasma_radial_thickness=200,
            outer_plasma_gap_radial_thickness=50,
            firstwall_radial_thickness=30,
            blanket_rear_wall_radial_thickness=30,
            number_of_tf_coils=16,
            rotation_angle=180,
            support_radial_thickness=90,
            inboard_blanket_radial_thickness=30,
            outboard_blanket_radial_thickness=30,
            elongation=2.00,
            triangularity=0.50,
            pf_coil_case_thicknesses=[10, 10, 10, 10],
            pf_coil_radial_thicknesses=[20, 50, 50, 20],
            pf_coil_vertical_thicknesses=[20, 50, 50, 20],
            pf_coil_radial_position=[500, 550, 550, 500],
            pf_coil_vertical_position=[270, 100, -100, -270],
            rear_blanket_to_tf_gap=50,
            outboard_tf_coil_radial_thickness=30,
            outboard_tf_coil_poloidal_thickness=30,
            divertor_position="lower"
        )

    def test_input_variable_names(self):
        """tests that the number of inputs variables is correct"""

        assert len(self.test_reactor.input_variables.keys()) == 29
        assert len(self.test_reactor.input_variable_names) == 29

    def test_single_null_submersion_tokamak_with_pf_and_tf_coils(self):
        """Creates a SingleNullSubmersionTokamak with pf and tf coils and checks
        that the correct number of components are created."""

        self.test_reactor.pf_coil_radial_thicknesses = [50, 50, 50, 50]
        self.test_reactor.pf_coil_vertical_thicknesses = [50, 50, 50, 50]
        self.test_reactor.pf_coil_to_tf_coil_radial_gap = 50
        self.test_reactor.pf_coil_case_thickness = 10
        self.test_reactor.outboard_tf_coil_radial_thickness = 100
        self.test_reactor.outboard_tf_coil_poloidal_thickness = 50
        self.test_reactor.tf_coil_to_rear_blanket_radial_gap = 20
        self.test_reactor.number_of_tf_coils = 16

        assert len(self.test_reactor.shapes_and_components) == 11

    def test_single_null_submersion_tokamak_rotation_angle_impacts_volume(
            self):
        """Creates SingleNullSubmersionTokamaks with different rotation angles and
        checks that the relative volumes of the components are correct."""

        self.test_reactor.rotation_angle = 90
        comps_90_vol = [
            comp.volume() for comp in self.test_reactor.shapes_and_components]
        self.test_reactor.rotation_angle = 180
        comps_180_vol = [
            comp.volume() for comp in self.test_reactor.shapes_and_components]

        for vol_90, vol_180 in zip(comps_90_vol, comps_180_vol):
            assert vol_90 == pytest.approx(vol_180 * 0.5, rel=0.1)

    def test_hash_value(self):
        """Creates a single null submersion reactor and checks that all shapes
        in the reactor are created when .shapes_and_components is first called.
        Checks that when .shapes_and_components is called again with no changes
        to the reactor, the shapes in the reactor are reconstructed and the
        previously constructed shapes are returned. Checks that when
        .shapes_and_components is called again with changes to the reactor,
        the shapes in the reactor are reconstructed and these new shapes are
        returned. Checks that the reactor_hash_value is only updated when the
        reactor is reconstructed."""

        self.test_reactor.pf_coil_radial_thicknesses = [30, 30, 30, 30]
        self.test_reactor.pf_coil_vertical_thicknesses = [30, 30, 30, 30]
        self.test_reactor.pf_coil_to_tf_coil_radial_gap = 50
        self.test_reactor.pf_coil_case_thickness = 10
        self.test_reactor.outboard_tf_coil_radial_thickness = 30
        self.test_reactor.outboard_tf_coil_poloidal_thickness = 30
        self.test_reactor.tf_coil_to_rear_blanket_radial_gap = 20
        self.test_reactor.number_of_tf_coils = 16

        assert self.test_reactor.reactor_hash_value is None
        for key in [
            "_inboard_tf_coils",
            "_center_column_shield",
            "_plasma",
            "_inboard_firstwall",
            "_inboard_blanket",
            "_firstwall",
            "_divertor",
            "_blanket",
            "_supports",
            "_outboard_rear_blanket_wall_upper",
            "_outboard_rear_blanket_wall_lower",
            "_outboard_rear_blanket_wall",
            "_tf_coil",
            "_pf_coil",
            "_pf_coils_casing"
        ]:
            assert key not in self.test_reactor.__dict__.keys()
        assert self.test_reactor.shapes_and_components is not None

        for key in [
            "_inboard_tf_coils",
            "_center_column_shield",
            "_plasma",
            "_inboard_firstwall",
            "_inboard_blanket",
            "_firstwall",
            "_divertor",
            "_blanket",
            "_supports",
            "_outboard_rear_blanket_wall_upper",
            "_outboard_rear_blanket_wall_lower",
            "_outboard_rear_blanket_wall",
            "_tf_coil",
            "_pf_coil",
            "_pf_coils_casing"
        ]:
            assert key in self.test_reactor.__dict__.keys()
        assert len(self.test_reactor.shapes_and_components) == 11
        assert self.test_reactor.reactor_hash_value is not None
        initial_hash_value = self.test_reactor.reactor_hash_value
        self.test_reactor.rotation_angle = 270
        assert self.test_reactor.reactor_hash_value == initial_hash_value
        assert self.test_reactor.shapes_and_components is not None
        assert self.test_reactor.reactor_hash_value != initial_hash_value




from typing import Optional, Tuple

from paramak import RotateStraightShape


class PoloidalFieldCoilCase(RotateStraightShape):
    """Creates a casing for a rectangular coil from inputs that
    describe the existing coil and the thickness of the casing required.

    Args:
        coil_height: the vertical (z axis) height of the coil (cm).
        coil_width: the horizontal (x axis) width of the coil (cm).
        center_point: the center of the coil (x,z) values (cm).
        casing_thickness: the thickness of the coil casing (cm).
    """

    def __init__(
        self,
        casing_thickness: Tuple[float, float],
        coil_height: float,
        coil_width: float,
        center_point: Tuple[float, float],
        color: Optional[Tuple[int, int, int]] = (1., 1., 0.498),
        **kwargs
    ) -> None:

        super().__init__(
            color=color,
            **kwargs
        )

        self.center_point = center_point
        self.height = coil_height
        self.width = coil_width
        self.casing_thickness = casing_thickness

    @property
    def center_point(self):
        return self._center_point

    @center_point.setter
    def center_point(self, value):
        self._center_point = value

    @property
    def height(self):
        return self._height

    @height.setter
    def height(self, height):
        self._height = height

    @property
    def width(self):
        return self._width

    @width.setter
    def width(self, width):
        self._width = width

    def find_points(self):
        """Finds the XZ points joined by straight connections that describe
        the 2D profile of the poloidal field coil case shape."""

        points = [
            (
                self.center_point[0] + self.width / 2.0,
                self.center_point[1] + self.height / 2.0,
            ),  # upper right
            (
                self.center_point[0] + self.width / 2.0,
                self.center_point[1] - self.height / 2.0,
            ),  # lower right
            (
                self.center_point[0] - self.width / 2.0,
                self.center_point[1] - self.height / 2.0,
            ),  # lower left
            (
                self.center_point[0] - self.width / 2.0,
                self.center_point[1] + self.height / 2.0,
            ),  # upper left
            (
                self.center_point[0] + self.width / 2.0,
                self.center_point[1] + self.height / 2.0,
            ),  # upper right
            (
                self.center_point[0] + \
                (self.casing_thickness + self.width / 2.0),
                self.center_point[1] + \
                (self.casing_thickness + self.height / 2.0),
            ),
            (
                self.center_point[0] + \
                (self.casing_thickness + self.width / 2.0),
                self.center_point[1] - \
                (self.casing_thickness + self.height / 2.0),
            ),
            (
                self.center_point[0] - \
                (self.casing_thickness + self.width / 2.0),
                self.center_point[1] - \
                (self.casing_thickness + self.height / 2.0),
            ),
            (
                self.center_point[0] - \
                (self.casing_thickness + self.width / 2.0),
                self.center_point[1] + \
                (self.casing_thickness + self.height / 2.0),
            ),
            (
                self.center_point[0] + \
                (self.casing_thickness + self.width / 2.0),
                self.center_point[1] + \
                (self.casing_thickness + self.height / 2.0),
            )
        ]

        self.points = points




import json
import os
import unittest
from pathlib import Path

import paramak
import pytest
from cadquery import Plane
from numpy.testing._private.utils import assert_


class TestShape(unittest.TestCase):

    def setUp(self):

        self.my_shape = paramak.CenterColumnShieldHyperbola(
            height=500,
            inner_radius=50,
            mid_radius=60,
            outer_radius=100,
        )

        self.test_rotate_mixed_shape = paramak.RotateMixedShape(
            rotation_angle=1,
            points=[
                (100, 0, "straight"),
                (200, 0, "circle"),
                (250, 50, "circle"),
                (200, 100, "straight"),
                (150, 100, "straight"),
                (140, 75, "straight"),
                (110, 45, "straight"),
            ]
        )
        self.test_extrude_mixed_shape = paramak.ExtrudeMixedShape(
            distance=1,
            points=[
                (100, 0, "straight"),
                (200, 0, "circle"),
                (250, 50, "circle"),
                (200, 100, "straight"),
                (150, 100, "straight"),
                (140, 75, "straight"),
                (110, 45, "straight"),
            ]
        )

    def test_graveyard_size_setting_type_checking(self):
        """Attempts to make a shape with a graveyard_size that is an float
        which should raise a ValueError"""

        def incorrect_graveyard_size_type():
            self.my_shape.graveyard_size = 'coucou'
        self.assertRaises(TypeError, incorrect_graveyard_size_type)

    def test_graveyard_size_setting_magnitude_checking(self):
        """Attempts to make a shape with a graveyard_size that is an int
        which should raise a ValueError"""

        def incorrect_graveyard_size_size():
            self.my_shape.graveyard_size = -10
        self.assertRaises(ValueError, incorrect_graveyard_size_size)

    def test_graveyard_error_when_no_offset_or_size(self):
        """Attempts to make a shape with a graveyard_size that is an int
        which should raise a ValueError"""

        def incorrect_graveyard():
            test_shape = paramak.RotateStraightShape(
                points=[(0, 0), (0, 20), (20, 20)],
                graveyard_size=None, graveyard_offset=None)
            test_shape.make_graveyard()

        self.assertRaises(ValueError, incorrect_graveyard)

    def test_show_runs_without_error(self):
        """checks that the jupyter notebook (with cadquery addition) runs
        without error."""

        self.test_extrude_mixed_shape.show()

    def test_make_graveyard_offset(self):
        """checks that the graveyard can be exported with the correct default
        parameters and that these parameters can be changed"""

        test_shape = paramak.RotateStraightShape(
            points=[(0, 0), (0, 20), (20, 20)],
            graveyard_size=None)
        os.system("rm graveyard.stp")

        test_shape.make_graveyard(graveyard_offset=50)
        graveyard_volume_1 = test_shape.graveyard.volume()

        test_shape.make_graveyard(graveyard_offset=200)
        graveyard_volume_2 = test_shape.graveyard.volume()

        assert graveyard_volume_2 > graveyard_volume_1

    def test_shape_default_properties(self):
        """Creates a Shape object and checks that the points attribute has
        a default of None."""

        test_shape = paramak.Shape()

        assert test_shape.points is None

    def test_azimuth_placement_angle_getting_setting(self):
        """Checks that the azimuth_placement_angle of a Shape can be
        changed to a single value or iterable."""

        test_shape = paramak.Shape()

        assert test_shape.azimuth_placement_angle == 0
        test_shape.azimuth_placement_angle = 180
        assert test_shape.azimuth_placement_angle == 180
        test_shape.azimuth_placement_angle = [0, 90, 180, 270]
        assert test_shape.azimuth_placement_angle == [0, 90, 180, 270]

    def test_incorrect_graveyard_offset_too_small(self):

        def incorrect_graveyard_offset_too_small():
            """Set graveyard_offset as a negative number which should raise an
            error"""

            self.my_shape.graveyard_offset = -3

        self.assertRaises(
            ValueError,
            incorrect_graveyard_offset_too_small
        )

    def test_incorrect_graveyard_offset_wrong_type(self):

        def incorrect_graveyard_offset_wrong_type():
            """Set graveyard_offset as a string which should raise an error"""

            self.my_shape.graveyard_offset = 'coucou'

        self.assertRaises(
            TypeError,
            incorrect_graveyard_offset_wrong_type
        )

    def test_missing_filename_arg_in_export_stp(self):
        """Checks that an error is raised when a stp export is requested
        without a filename."""

        def incorrect_args():
            self.my_shape.export_stp()

        self.assertRaises(
            TypeError,
            incorrect_args
        )

    def test_missing_filename_arg_in_export_stl(self):
        """Checks that an error is raised when a stl export is requested
        without a filename."""

        def incorrect_args():
            self.my_shape.export_stl()

        self.assertRaises(
            TypeError,
            incorrect_args
        )

    def test_incorrect_color_values(self):
        """Checks that an error is raised when the color of a shape is
        defined as an invalid string."""

        def incorrect_color_string():
            paramak.Shape(color=('1', '0', '1'))

        self.assertRaises(
            ValueError,
            incorrect_color_string
        )

    def test_incorrect_workplane(self):
        """Creates Shape object with incorrect workplane and checks errors
        are raised."""

        with pytest.raises(ValueError):
            paramak.Shape(workplane="AB")

        with pytest.raises(TypeError):
            paramak.Shape(workplane=2)

        with pytest.raises(TypeError):
            paramak.Shape(workplane=[1, 2])

    def test_incorrect_points(self):
        """Creates Shape objects and checks errors are raised correctly when
        specifying points."""

        test_shape = paramak.Shape()

        def incorrect_points_end_point_is_start_point():
            """Checks ValueError is raised when the start and end points are
            the same."""
            # setting straight otherwise another error is caught
            test_shape.connection_type = "straight"
            test_shape.points = [(0, 200), (200, 100), (0, 0), (0, 200)]

        # check that an error is raised
        with pytest.raises(ValueError) as err:
            incorrect_points_end_point_is_start_point()

        # check that the correct error was raised
        expected_err_message = "The coordinates of the last and first points are"
        assert expected_err_message in str(err.value)

        def incorrect_points_missing_z_value():
            """Checks ValueError is raised when a point is missing a z
            value."""

            test_shape.points = [(0, 200), (200), (0, 0), (0, 50)]

        self.assertRaises(ValueError, incorrect_points_missing_z_value)

        def incorrect_points_not_a_list():
            """Checks ValueError is raised when the points are not a list."""

            test_shape.points = "(0, 0), (0, 20), (20, 20), (20, 0)"

        self.assertRaises(ValueError, incorrect_points_not_a_list)

        def incorrect_points_wrong_number_of_entries():
            """Checks ValueError is raised when individual points dont have 2
            or 3 entries."""

            test_shape.points = [(0, 0), (0, 20), (20, 20, 20, 20)]

        self.assertRaises(ValueError, incorrect_points_wrong_number_of_entries)

        def incorrect_x_point_value_type():
            """Checks ValueError is raised when X point is not a number."""

            test_shape.points = [("string", 0), (0, 20), (20, 20)]

        self.assertRaises(ValueError, incorrect_x_point_value_type)

        def incorrect_y_point_value_type():
            """Checks ValueError is raised when Y point is not a number."""

            test_shape.points = [(0, "string"), (0, 20), (20, 20)]

        self.assertRaises(ValueError, incorrect_y_point_value_type)

    def test_create_limits(self):
        """Creates a Shape object and checks that the create_limits function
        returns the expected values for x_min, x_max, z_min and z_max."""

        test_shape = paramak.Shape(connection_type='straight')

        test_shape.points = [
            (0, 0),
            (0, 10),
            (0, 20),
            (10, 20),
            (20, 20),
            (20, 10),
            (20, 0),
            (10, 0),
        ]

        assert test_shape.create_limits() == (0.0, 20.0, 0.0, 20.0)

        # test with a component which has a find_points method
        test_shape2 = paramak.Plasma()
        test_shape2.create_limits()
        assert test_shape2.x_min is not None

    def test_create_limits_error(self):
        """Checks error is raised when no points are given."""

        test_shape = paramak.Shape()

        def limits():
            test_shape.create_limits()
        self.assertRaises(ValueError, limits)

    def test_initial_solid_construction(self):
        """Creates a shape and checks that a cadquery solid with a unique hash
        value is created when .solid is called."""

        test_shape = paramak.RotateStraightShape(
            points=[(0, 0), (0, 20), (20, 20), (20, 0)], rotation_angle=360
        )

        assert test_shape.hash_value is None
        assert test_shape.solid is not None
        assert type(test_shape.solid).__name__ == "Workplane"
        assert test_shape.hash_value is not None

    def test_solid_return(self):
        """Checks that the same cadquery solid with the same unique hash value
        is returned when shape.solid is called again after no changes have been
        made to the Shape."""

        test_shape = paramak.RotateStraightShape(
            points=[(0, 0), (0, 20), (20, 20), (20, 0)], rotation_angle=360
        )

        assert test_shape.solid is not None
        initial_hash_value = test_shape.hash_value
        assert test_shape.solid is not None
        assert initial_hash_value == test_shape.hash_value

    def test_conditional_solid_reconstruction(self):
        """Checks that a new cadquery solid with a new unique hash value is
        constructed when shape.solid is called after changes to the Shape have
        been made."""

        test_shape = paramak.RotateStraightShape(
            points=[(0, 0), (0, 20), (20, 20)], rotation_angle=360
        )

        assert test_shape.solid is not None
        assert test_shape.hash_value is not None
        initial_hash_value = test_shape.hash_value

        test_shape.rotation_angle = 180

        assert test_shape.solid is not None
        assert test_shape.hash_value is not None
        assert initial_hash_value != test_shape.hash_value

    def test_hash_value_update(self):
        """Checks that the hash value of a Shape is not updated until a new
        cadquery solid has been created."""

        test_shape = paramak.RotateStraightShape(
            points=[(0, 0), (0, 20), (20, 20)], rotation_angle=360
        )
        test_shape.solid
        assert test_shape.hash_value is not None
        initial_hash_value = test_shape.hash_value

        test_shape.rotation_angle = 180
        assert test_shape.hash_value == initial_hash_value
        test_shape.solid
        assert test_shape.hash_value != initial_hash_value

    def test_export_html(self):
        """Checks a plotly figure of the Shape is exported by the export_html
        method with the correct filename with RGB and RGBA colors."""

        test_shape = paramak.RotateStraightShape(
            points=[(0, 0), (0, 20), (20, 20), (20, 0)], rotation_angle=360
        )

        os.system("rm filename.html")
        test_shape.export_html('filename')
        assert Path("filename.html").exists() is True
        os.system("rm filename.html")
        test_shape.color = (1, 0, 0, 0.5)
        test_shape.export_html('filename')
        assert Path("filename.html").exists() is True
        os.system("rm filename.html")

    def test_export_html_view_planes(self):
        """Checks a plotly figure of the Shape is exported by the export_html
        method with a range of different view_plane options."""

        test_shape = paramak.RotateStraightShape(
            points=[(0, 0), (0, 20), (20, 20), (20, 0)], rotation_angle=180
        )

        for view_plane in ['XZ', 'XY', 'YZ', 'YX', 'ZY', 'ZX', 'RZ', 'XYZ']:
            os.system("rm *.html")
            test_shape.export_html(
                filename='filename',
                view_plane=view_plane
            )
            assert Path("filename.html").exists() is True

    def test_export_html_with_points_None(self):
        """Checks that an error is raised when points is None and export_html
        """
        test_shape = paramak.Shape()

        def export():
            test_shape.export_html("out.html")
        self.assertRaises(ValueError, export)

    def test_export_html_with_wire_None(self):
        """Checks that an error is raised when wire is None and export_html
        """
        test_shape = paramak.Shape(
            points=[(0, 0), (0, 20), (20, 20), (20, 0)],
            connection_type='straight',
        )
        test_shape.wire = None

        def export():
            test_shape.export_html("out.html")
        self.assertRaises(ValueError, export)

    def test_export_3d_html(self):
        """Checks the 3d html file is exported by the export_html_3d method
        with the correct filename"""

        test_shape = paramak.RotateStraightShape(
            points=[(0, 0), (0, 20), (20, 20), (20, 0)], rotation_angle=360
        )

        os.system("rm filename.html")
        filename = test_shape.export_html_3d('filename.html')
        if filename is not None:
            assert Path("filename.html").exists() is True

    def test_invalid_stp_filename(self):
        """Checks ValueError is raised when invalid stp filenames are used."""

        def invalid_filename_suffix():

            test_shape = paramak.RotateStraightShape(
                points=[(0, 0), (0, 20), (20, 20)],
            )
            test_shape.export_stp(filename='test_shape.txt')

        self.assertRaises(ValueError, invalid_filename_suffix)

        def invalid_filename_type():

            test_shape = paramak.RotateStraightShape(
                points=[(0, 0), (0, 20), (20, 20)],
            )
            test_shape.export_stp(filename=1)

        self.assertRaises(TypeError, invalid_filename_type)

    def test_invalid_stl_filename(self):
        """Checks ValueError is raised when invalid stl filenames are used."""

        def invalid_filename_suffix():

            test_shape = paramak.RotateStraightShape(
                points=[(0, 0), (0, 20), (20, 20)],
            )
            test_shape.export_stl(filename='test_shape.txt')

        self.assertRaises(ValueError, invalid_filename_suffix)

        def invalid_filename_type():

            test_shape = paramak.RotateStraightShape(
                points=[(0, 0), (0, 20), (20, 20)],
            )
            test_shape.export_stp(filename=1)

        self.assertRaises(TypeError, invalid_filename_type)

    def test_invalid_color(self):
        """Checks ValueError is raised when invalid colors are used."""

        def invalid_color_type():

            paramak.RotateStraightShape(
                points=[(0, 0), (0, 20), (20, 20)],
                color=255
            )

        self.assertRaises(ValueError, invalid_color_type)

        def invalid_color_length():

            paramak.RotateStraightShape(
                points=[(0, 0), (0, 20), (20, 20)],
                color=(255, 255, 255, 1, 1)
            )

        self.assertRaises(ValueError, invalid_color_length)

    def test_volumes_add_up_to_total_volume_Compound(self):
        """Checks the volume and volumes attributes are correct types
        and that the volumes sum to equalt the volume for a Compound."""

        test_shape = paramak.PoloidalFieldCoilSet(
            heights=[10, 10],
            widths=[20, 20],
            center_points=[(15, 15), (50, 50)]
        )

        assert isinstance(test_shape.volume(), float)
        assert isinstance(test_shape.volume(split_compounds=True), list)
        assert isinstance(test_shape.volume(split_compounds=True)[0], float)
        assert isinstance(test_shape.volume(split_compounds=True)[1], float)
        assert len(test_shape.volume(split_compounds=True)) == 2
        assert sum(
            test_shape.volume(
                split_compounds=True)) == pytest.approx(
            test_shape.volume())

    def test_volumes_add_up_to_total_volume(self):
        """Checks the volume and volumes attributes are correct types
        and that the volumes sum to equalt the volume."""

        test_shape = paramak.PoloidalFieldCoil(
            center_point=(100, 100),
            height=50,
            width=50
        )

        assert isinstance(test_shape.volume(), float)
        assert isinstance(test_shape.volume(split_compounds=True), list)
        assert isinstance(test_shape.volume(split_compounds=True)[0], float)
        assert len(test_shape.volume(split_compounds=True)) == 1
        assert sum(
            test_shape.volume(
                split_compounds=True)) == pytest.approx(
            test_shape.volume())

    def test_areas_add_up_to_total_area_Compound(self):
        """Checks the area and areas attributes are correct types
        and that the areas sum to equalt the area for a Compound."""

        test_shape = paramak.PoloidalFieldCoilSet(
            heights=[10, 10],
            widths=[20, 20],
            center_points=[(15, 15), (50, 50)]
        )

        assert isinstance(test_shape.area, float)
        assert isinstance(test_shape.areas, list)
        assert isinstance(test_shape.areas[0], float)
        assert isinstance(test_shape.areas[1], float)
        assert isinstance(test_shape.areas[2], float)
        assert isinstance(test_shape.areas[3], float)
        assert isinstance(test_shape.areas[4], float)
        assert isinstance(test_shape.areas[5], float)
        assert isinstance(test_shape.areas[6], float)
        assert isinstance(test_shape.areas[7], float)
        assert len(test_shape.areas) == 8
        assert sum(test_shape.areas) == pytest.approx(test_shape.area)

    def test_areas_add_up_to_total_area(self):
        """Checks the area and areas attributes are correct types
        and that the areas sum to equalt the area."""

        test_shape = paramak.PoloidalFieldCoil(
            center_point=(100, 100),
            height=50,
            width=50
        )

        assert isinstance(test_shape.area, float)
        assert isinstance(test_shape.areas, list)
        assert isinstance(test_shape.areas[0], float)
        assert isinstance(test_shape.areas[1], float)
        assert isinstance(test_shape.areas[2], float)
        assert isinstance(test_shape.areas[3], float)
        assert len(test_shape.areas) == 4
        assert sum(test_shape.areas) == pytest.approx(test_shape.area)

    def test_create_patch_error(self):
        """Checks _create_patch raises a ValueError when points is None."""

        test_shape = paramak.Shape()

        def patch():
            test_shape._create_patch()
        self.assertRaises(ValueError, patch)

    def test_create_patch_alpha(self):
        """Checks _create_patch returns a patch when alpha is given."""

        test_shape = paramak.PoloidalFieldCoil(
            center_point=(100, 100),
            height=50,
            width=50,
            color=(0.5, 0.5, 0.5, 0.1)
        )
        assert test_shape._create_patch() is not None

    def test_azimuth_placement_angle_error(self):
        """Checks an error is raised when invalid value for
        azimuth_placement_angle is set.
        """

        test_shape = paramak.Shape()

        def angle_str():
            test_shape.azimuth_placement_angle = "coucou"

        def angle_str_in_Iterable():
            test_shape.azimuth_placement_angle = [0, "coucou"]

        self.assertRaises(ValueError, angle_str)
        self.assertRaises(ValueError, angle_str_in_Iterable)

    def test_name_error(self):
        """Checks an error is raised when invalid value for name is set."""

        test_shape = paramak.Shape()

        def name_float():
            test_shape.name = 2.0

        def name_int():
            test_shape.name = 1

        def name_list():
            test_shape.name = ['coucou']

        self.assertRaises(ValueError, name_float)
        self.assertRaises(ValueError, name_int)
        self.assertRaises(ValueError, name_list)

    def test_tet_mesh_error(self):
        """Checks an error is raised when invalid value for tet_mesh is set.
        """

        test_shape = paramak.Shape()

        def tet_mesh_float():
            test_shape.tet_mesh = 2.0

        def tet_mesh_int():
            test_shape.tet_mesh = 1

        def tet_mesh_list():
            test_shape.tet_mesh = ['coucou']

        self.assertRaises(ValueError, tet_mesh_float)
        self.assertRaises(ValueError, tet_mesh_int)
        self.assertRaises(ValueError, tet_mesh_list)

    def test_workplane_of_type_cadquery_plane(self):
        """Tests that a Cadquery.Plane is accepted as a workplane entry"""

        normal_vec = (1, 1, 1)
        workplane = Plane(origin=(0, 0, 0), xDir=(-1, 1, 0), normal=normal_vec)
        # in future releases of CQ, origin and xDir will be optional

        test_shape = paramak.Shape(
            workplane=workplane,
        )

        assert isinstance(test_shape.workplane, Plane)

    def test_get_rotation_axis(self):
        """Creates a shape and test the expected rotation_axis is the correct
        values for several cases
        """
        shape = paramak.Shape()
        expected_dict = {
            "X": [(-1, 0, 0), (1, 0, 0)],
            "-X": [(1, 0, 0), (-1, 0, 0)],
            "Y": [(0, -1, 0), (0, 1, 0)],
            "-Y": [(0, 1, 0), (0, -1, 0)],
            "Z": [(0, 0, -1), (0, 0, 1)],
            "-Z": [(0, 0, 1), (0, 0, -1)],
        }
        # test with axis from string
        for axis in expected_dict:
            shape.rotation_axis = axis
            assert shape.get_rotation_axis()[0] == expected_dict[axis]
            assert shape.get_rotation_axis()[1] == axis

        # test with axis from list of two points
        expected_axis = [(-1, -2, -3), (1, 4, 5)]
        shape.rotation_axis = expected_axis
        assert shape.get_rotation_axis()[0] == expected_axis
        assert shape.get_rotation_axis()[1] == "custom_axis"

        # test with axis from workplane
        shape.rotation_axis = None

        workplanes = ["XY", "XZ", "YZ"]
        expected_axis = ["Y", "Z", "Z"]
        for wp, axis in zip(workplanes, expected_axis):
            shape.workplane = wp
            assert shape.get_rotation_axis()[0] == expected_dict[axis]
            assert shape.get_rotation_axis()[1] == axis

        # test with axis from path_workplane
        for wp, axis in zip(workplanes, expected_axis):
            shape.path_workplane = wp
            assert shape.get_rotation_axis()[0] == expected_dict[axis]
            assert shape.get_rotation_axis()[1] == axis

    def test_rotation_axis_error(self):
        """Checks errors are raised when incorrect values of rotation_axis are
        set
        """
        incorrect_values = [
            "coucou",
            2,
            2.2,
            [(1, 1, 1), 'coucou'],
            [(1, 1, 1), 1],
            [(1, 1, 1), 1.0],
            [(1, 1, 1), (1, 1, 1)],
            [(1, 1, 1), (1, 0, 1, 2)],
            [(1, 1, 1, 2), (1, 0, 2)],
            [(1, 1, 2), [1, 0, 2]],
            [(1, 1, 1)],
            [(1, 1, 1), (1, 'coucou', 1)],
            [(1, 1, 1), (1, 0, 1), (1, 2, 3)],
        ]
        shape = paramak.Shape()

        def set_value():
            shape.rotation_axis = incorrect_values[i]

        for i in range(len(incorrect_values)):
            self.assertRaises(ValueError, set_value)

    def test_setting_color_incorrectly_too_large(self):
        """Sets the shape.colour outside of the the 0 to 1 range"""

        def check_correct_error_is_rasied():
            self.my_shape.color = (255, 255, 2)

        self.assertRaises(ValueError, check_correct_error_is_rasied)

    def test_setting_color_incorrectly_too_small(self):
        """Sets the shape.colour outside of the the 0 to 1 range"""

        def check_correct_error_is_rasied():
            self.my_shape.color = (-1, 0, 0)

        self.assertRaises(ValueError, check_correct_error_is_rasied)

    def test_reuse_points_between_shapes(self):
        """Checks that points can be reused between shapes"""
        points = [
            (100, 0, "straight"),
            (200, 0, "straight"),
            (250, 50, "straight"),
            (200, 100, "straight"),
        ]
        paramak.Shape(
            points=points
        )
        paramak.Shape(
            points=points
        )

    def test_reuse_points_and_connections(self):
        """Checks that points can be reused between shapes"""
        points = [
            (100, 0, "straight"),
            (200, 0, "straight"),
            (250, 50, "straight"),
            (200, 100, "straight"),
        ]
        test_shape = paramak.Shape(
            points=points
        )

        assert test_shape.points == [
            (100, 0, "straight"),
            (200, 0, "straight"),
            (250, 50, "straight"),
            (200, 100, "straight"),
        ]

    def test_reuse_points(self):
        """Checks that points can be reused between shapes"""
        points = [
            (100, 0),
            (200, 0),
            (250, 50),
            (200, 100),
        ]
        test_shape = paramak.Shape(
            points=points,
            connection_type='straight'
        )

        assert test_shape.points == [
            (100, 0),
            (200, 0),
            (250, 50),
            (200, 100),
        ]


if __name__ == "__main__":
    unittest.main()




import unittest

import paramak


class TestInnerTfCoilsCircular(unittest.TestCase):

    def setUp(self):
        self.test_shape = paramak.InnerTfCoilsCircular(
            height=500,
            inner_radius=50,
            outer_radius=150,
            number_of_coils=6,
            gap_size=5
        )

    def test_default_parameters(self):
        """Checks that the default parameters of an InnerTfCoilsCircular are correct."""

        assert self.test_shape.rotation_angle == 360
        assert self.test_shape.azimuth_start_angle == 0
        assert self.test_shape.workplane == "XY"
        assert self.test_shape.rotation_axis == "Z"

    def test_points_calculation(self):
        """Checks that the points used to construct the InnerTFCoilsCircular
        component are calculated correctly from the parameters given."""

        assert self.test_shape.points == [
            (49.937460888595446, 2.5, 'circle'),
            (43.300748759659555, 25.000903120744287, 'circle'),
            (27.1320420790315, 41.99824154201773, 'straight'),
            (77.154447582418, 128.6358861991937, 'circle'),
            (129.90375269002172, 75.00010024693078, 'circle'),
            (149.97916521970643, 2.5, 'straight'),
        ]

    def test_processed_points_calculation(self):
        """Checks that the processed_points used to construct the
        InnerTFCoilsCircular component are calculated correctly from the
        parameters given."""

        assert self.test_shape.processed_points == [
            (49.937460888595446, 2.5, 'circle'),
            (43.300748759659555, 25.000903120744287, 'circle'),
            (27.1320420790315, 41.99824154201773, 'straight'),
            (77.154447582418, 128.6358861991937, 'circle'),
            (129.90375269002172, 75.00010024693078, 'circle'),
            (149.97916521970643, 2.5, 'straight'),
            (49.937460888595446, 2.5, 'circle')
        ]

    def test_creation(self):
        """Creates an inner tf coil using the InnerTfCoilsCircular parametric
        component and checks that a cadquery solid is created."""

        assert self.test_shape.solid is not None
        assert self.test_shape.volume() > 1000

    def test_azimuth_offset(self):
        """Creates an inner tf coil using the InnerTfCoilsCircular parametric
        component and checks that the azimuthal start angle can be changed
        correctly."""

        assert self.test_shape.azimuth_placement_angle == [
            0, 60, 120, 180, 240, 300]
        self.test_shape.azimuth_start_angle = 20
        assert self.test_shape.azimuth_placement_angle == [
            20, 80, 140, 200, 260, 320]

    def test_attributes(self):
        """Checks that changing the attributes of InnerTfCoilsCircular affects
        the cadquery solid produced."""

        test_volume = self.test_shape.volume()

        self.test_shape.height = 1000
        assert test_volume == self.test_shape.volume() * 0.5
        self.test_shape.height = 500
        self.test_shape.inner_radius = 30
        assert test_volume < self.test_shape.volume()
        self.test_shape.inner_radius = 50
        self.test_shape.outer_radius = 170
        assert test_volume < self.test_shape.volume()

    def test_gap_size(self):
        """Checks that a ValueError is raised when a too large gap_size is
        used."""

        def test_incorrect_gap_size():
            self.test_shape.inner_radius = 20
            self.test_shape.outer_radius = 40
            self.test_shape.gap_size = 50
            self.test_shape.solid

        self.assertRaises(
            ValueError,
            test_incorrect_gap_size
        )




from typing import Optional, Tuple

from cadquery import Workplane
from paramak import Shape


class RotateCircleShape(Shape):
    """Rotates a circular 3d CadQuery solid from a central point and a radius

    Args:
        radius: radius of the shape
        rotation_angle: The rotation_angle to use when revolving the solid
            (degrees). Defaults to 360.0.
    """

    def __init__(
        self,
        radius: float,
        rotation_angle: Optional[float] = 360.0,
        color: Optional[Tuple[float, float, float, Optional[float]]] = (1., 1., 0.6),
        **kwargs
    ):

        super().__init__(
            color=color,
            **kwargs
        )
        self.radius = radius
        self.rotation_angle = rotation_angle

    @property
    def rotation_angle(self):
        return self._rotation_angle

    @rotation_angle.setter
    def rotation_angle(self, value):
        self._rotation_angle = value

    @property
    def radius(self):
        return self._radius

    @radius.setter
    def radius(self, value):
        self._radius = value

    def create_solid(self):
        """Creates a rotated 3d solid using points with circular edges.

           Returns:
              A CadQuery solid: A 3D solid volume
        """

        wire = (
            Workplane(self.workplane)
            .moveTo(self.points[0][0], self.points[0][1])
            .circle(self.radius)
        )

        self.wire = wire

        solid = wire.revolve(self.rotation_angle)

        solid = self.rotate_solid(solid)
        solid = self.perform_boolean_operations(solid)
        self.solid = solid
        return solid



from typing import Optional, Tuple, Union

import cadquery as cq
import numpy as np
from paramak.parametric_shapes.extruded_mixed_shape import ExtrudeMixedShape
from paramak.utils import calculate_wedge_cut


class ToroidalFieldCoilRectangleRoundCorners(ExtrudeMixedShape):
    """Creates geometry for TF coil with rounded corners. Finds the coordinates
    for vertices of a TF coil, in a 2D profile on the XZ plane using the main
    function find_points() which takes 3 positional arguments for the TF coil
    parameters, and takes three additional boolean arguments.

    Arguments:
        lower_inner_coordinates (Tuple): the (X,Z) coordinate of the inner
            corner of the lower end of the coil (cm)
        mid_point_coordinates (Tuple): the (X,Z) coordinate of the mid
            point of the vertical section (cm)
        thickness: The thickness in the (X,Z) plane of the toroidal
        field coils (cm)
        extrusiondistance: The total extruded thickness of the coils
            when in the y-direction (centered extrusion)
        coil_count: The number of coils placed in the model
            (changing azimuth_placement_angle by dividing 360 by the amount
            given). Defaults to 1
        with_inner_leg: Boolean to include the inside of the Coils
            defaults to False
    """

    def __init__(
        self,
        lower_inner_coordinates: Tuple[float, float],
        mid_point_coordinates: Tuple[float, float],
        thickness: Union[float, int],
        distance: float,
        number_of_coils: int = 1,
        with_inner_leg: Optional[bool] = False,
        **kwargs
    ) -> None:

        super().__init__(
            distance=distance,
            **kwargs
        )

        self._lower_inner_coordinates = lower_inner_coordinates
        self._mid_point_coordinates = mid_point_coordinates
        self._thickness = thickness
        self._distance = distance
        self.number_of_coils = number_of_coils
        self._with_inner_leg = with_inner_leg
        self._inner_leg_connection_points = []
        self._analyse_attributes = [
            0,
            0,
            0,
            0
        ]
        self._base_length = 0
        self._height = 0
        self._inner_curve_radius = 0
        self._outter_curve_radius = 0

        if len(lower_inner_coordinates) != 2 or len(
                mid_point_coordinates) != 2:
            msg = ('The input tuples are too long or too short, they must be '
                   '2 element long')
            raise ValueError(msg)

        if self._lower_inner_coordinates[0] > self._mid_point_coordinates[0]:
            raise ValueError(
                "The middle point's x-coordinate must be larger than the lower",
                "inner point's x-coordinate")

    def _find_base_and_height(self):
        # Adding hidden attributes for analyse list population
        # inner base length of the coil
        self._base_length = self._mid_point_coordinates[0] - \
            self._lower_inner_coordinates[0]
        self._analyse_attributes[0] = self._base_length

        # height of the coil
        self._height = abs(
            self.mid_point_coordinates[1] - self.lower_inner_coordinates[1]) * 2
        self._analyse_attributes[1] = self._height

    def _find_radii(self):
        # Inner and outter radius of curvature for the corners
        # The inner curvature is scales as a function of the base length
        # of the coil and its thickness as long as the thickness does not exceed the base length
        # if the thickness/base length ratio is larger or equal to 1
        # it takes 10% of the thickness as the inner curve radius
        # this to avoid having coordinates before the previous or at the same spot as Paramak
        # cannot compute it
        self._find_base_and_height()
        if self._thickness / self._base_length >= 1:
            self._inner_curve_radius = self._thickness * 0.1
            self._outter_curve_radius = self._thickness * 1.1
        else:
            self._outter_curve_radius = (
                1 + (self._thickness / self._base_length)) * self._thickness
            self._inner_curve_radius = (self._thickness**2) / self._base_length

        self._analyse_attributes[2] = self._inner_curve_radius
        self._analyse_attributes[3] = self._outter_curve_radius

    @property
    def azimuth_placement_angle(self):
        self.find_azimuth_placement_angle()
        return self._azimuth_placement_angle

    @azimuth_placement_angle.setter
    def azimuth_placement_angle(self, val):
        self._azimuth_placement_angle = val

    @property
    def lower_inner_coordinates(self):
        return self._lower_inner_coordinates

    @lower_inner_coordinates.setter
    def lower_inner_coordinates(self, val):
        if not isinstance(val, tuple):
            raise TypeError("Input Coordinate must be a tuple!")
        if len(val) != 2:
            raise ValueError("Input Tuple must be 2 elements long!")
        if not isinstance(val[0], (float, int)):
            raise TypeError("Input X Coordinates must be a number!")
        if not isinstance(val[1], (float, int)):
            raise TypeError("Input Z Coordinates must be a number!")
        if val[0] > self._mid_point_coordinates[0]:
            raise ValueError(
                "Mid Point's x-coordinate, must be larger than lower point's!")
        self._lower_inner_coordinates = val

    @property
    def mid_point_coordinates(self):
        return self._mid_point_coordinates

    @mid_point_coordinates.setter
    def mid_point_coordinates(self, val):
        if not isinstance(val, tuple):
            raise TypeError("Input Coordinate must be a tuple!")
        if len(val) != 2:
            raise ValueError("Input Tuple must be 2 elements long!")
        if not isinstance(val[0], (float, int)):
            raise TypeError("Input X Coordinates must be a number!")
        if not isinstance(val[1], (float, int)):
            raise TypeError("Input Z Coordinates must be a number!")
        if val[0] < self._lower_inner_coordinates[0]:
            raise ValueError(
                "Mid Point's x-coordinate, must be larger than lower point's!")
        self._mid_point_coordinates = val

    @property
    def thickness(self):
        return self._thickness

    @thickness.setter
    def thickness(self, val):
        if not isinstance(val, (float, int)):
            raise TypeError("Input Thickness must be a number!")
        self._find_radii()
        self._thickness = val

    @property
    def distance(self):
        return self._distance

    @distance.setter
    def distance(self, val):
        if not isinstance(val, (float, int)):
            raise TypeError("Input Distance must be a number!")
        self._distance = val

    @property
    def number_of_coils(self):
        return self._number_of_coils

    @number_of_coils.setter
    def number_of_coils(self, val):
        if not isinstance(val, int):
            raise TypeError("Input Distance must be an integer number!")
        self._number_of_coils = val

    @property
    def analyse_attributes(self):
        self.find_points()
        return self._analyse_attributes

    @property
    def with_inner_leg(self):
        return self._with_inner_leg

    @with_inner_leg.setter
    def with_inner_leg(self, val):
        if not isinstance(val, bool):
            raise TypeError("With Inner Leg must be True or False")

    def find_points(self):
        """
        lower_inner_coordinates must be a 2 element tuple
        mid_point_coordinates must be a 2 elemenet tuple
        thickness must be a float or an int
        """
        self._find_radii()

        lower_x, lower_z = self._lower_inner_coordinates
        mid_x, mid_z = self._mid_point_coordinates

        # redifine values to be floats to make it look consistent
        lower_x, lower_z, mid_x, mid_z, thickness = float(lower_x), float(
            lower_z), float(mid_x), float(mid_z), float(self._thickness)

        # Define differences to avoid miss claculation due to signs
        base_length = self._analyse_attributes[0]
        height = self._analyse_attributes[1]

        # 10 points/tuples for initial calculation and to get aux points
        point1 = (lower_x, lower_z)
        point2 = (point1[0] + base_length, point1[1])
        point3 = (point2[0], point2[1] + height)
        point4 = (point1[0], point1[1] + height)
        point5 = (point4[0], point4[1] + thickness)
        point6 = (point3[0], point4[1] + thickness)
        #point7 = (point3[0] + thickness, point3[1])
        point8 = (point2[0] + thickness, point2[1])
        #point9 = (point2[0], point2[1] - thickness)
        point10 = (lower_x, lower_z - thickness)

        inner_curve_radius = self._analyse_attributes[2]
        outter_curve_radius = self._analyse_attributes[3]

        #  New subroutines to calculate inner and outter curve mid-points,    #
        #  x and y displacement from existing points                          #
        #  long shift does a sin(45)*radius of curvature shift                #
        #  short shift does a (1-sin(45))*radius of curvature shift           #

        def shift_long(radius):
            """radius is the radius of curvature"""
            return (2**0.5) * 0.5 * radius

        def shift_short(radius):
            """radius is the radius of curvature"""
            return (2 - (2**0.5)) * 0.5 * radius

        point11 = (point2[0] - inner_curve_radius, point2[1])
        point12 = (point11[0] + shift_long(inner_curve_radius),
                   point11[1] + shift_short(inner_curve_radius))
        point13 = (point2[0], point2[1] + inner_curve_radius)
        point14 = (point3[0], point3[1] - inner_curve_radius)
        point15 = (point14[0] - shift_short(inner_curve_radius),
                   point14[1] + shift_long(inner_curve_radius))
        point16 = (point3[0] - inner_curve_radius, point3[1])
        point17 = (point6[0] - inner_curve_radius, point6[1])
        point18 = (point17[0] + shift_long(outter_curve_radius),
                   point17[1] - shift_short(outter_curve_radius))
        point19 = (point14[0] + thickness, point14[1])
        point20 = (point8[0], point8[1] + inner_curve_radius)
        point21 = (point18[0], point20[1] - shift_long(outter_curve_radius))
        point22 = (point11[0], point11[1] - thickness)

        # List holding the points that are being returned by the function
        points = [
            point1,
            point11,
            point12,
            point13,
            point14,
            point15,
            point16,
            point4,
            point5,
            point17,
            point18,
            point19,
            point20,
            point21,
            point22,
            point10]
        # List that holds the points with the corresponding line types
        tri_points = []
        lines = ["straight"] + ['circle'] * 2 + ['straight'] \
            + ['circle'] * 2 + ['straight'] * 3 + ['circle'] * 2 \
            + ['straight'] + ['circle'] * 2 + ['straight'] * 2

        for i in enumerate(points):
            tri_points.append(points[i[0]] + (lines[i[0]],))

        self.points = tri_points

        inner_point1 = (point1[0], point1[1])
        inner_point2 = (point1[0] + thickness, point1[1])
        inner_point3 = (point4[0] + thickness, point4[1])
        inner_point4 = (point4[0], point4[1])

        self._inner_leg_connection_points = [
            inner_point1, inner_point2, inner_point3, inner_point4]

        return tri_points

    def find_azimuth_placement_angle(self):
        """ Finds the placement angles from the number of coils
            given in a 360 degree """

        angles = list(
            np.linspace(
                0,
                360,
                self._number_of_coils,
                endpoint=False))
        self.azimuth_placement_angle = angles

    def create_solid(self):
        """ Creates a Cadquery 3D geometry

        Returns:
            CadQuery solid: A 3D solid Volume """

        # Create solid from points
        points = [ps[:2] for ps in self.processed_points]

        wire = cq.Workplane(self.workplane).moveTo(points[0][0], points[0][1]) \
            .lineTo(points[1][0], points[1][1]) \
            .threePointArc((points[2][0], points[2][1]), (points[3][0], points[3][1])) \
            .lineTo(points[4][0], points[4][1]) \
            .threePointArc((points[5][0], points[5][1]), (points[6][0], points[6][1])) \
            .lineTo(points[7][0], points[7][1]) \
            .lineTo(points[8][0], points[8][1]) \
            .lineTo(points[9][0], points[9][1]) \
            .threePointArc((points[10][0], points[10][1]), (points[11][0], points[11][1])) \
            .lineTo(points[12][0], points[12][1]) \
            .threePointArc((points[13][0], points[13][1]), (points[14][0], points[14][1])) \
            .lineTo(points[15][0], points[15][1]) \
            .lineTo(points[16][0], points[16][1]).close().consolidateWires()

        solid = wire.extrude(distance=-self._distance / 2, both=True)
        solid = self.rotate_solid(solid)

        cutting_wedge = calculate_wedge_cut(self)
        solid = self.perform_boolean_operations(solid, wedge_cut=cutting_wedge)

        self.solid = solid

        if self._with_inner_leg:
            inner_leg_solid = cq.Workplane(self.workplane)
            inner_leg_solid = inner_leg_solid.polyline(
                self._inner_leg_connection_points).close().extrude(
                distance=-self._distance / 2, both=True)

            inner_leg_solid = self.rotate_solid(inner_leg_solid)
            inner_leg_solid = self.perform_boolean_operations(
                inner_leg_solid, wedge_cut=cutting_wedge)

            solid = cq.Compound.makeCompound(
                [a.val() for a in [inner_leg_solid, solid]])

            self.solid = solid

        return solid




from typing import Optional, Tuple
from paramak import RotateStraightShape


class PoloidalFieldCoil(RotateStraightShape):
    """Creates a rectangular poloidal field coil.

    Args:
        height: the vertical (z axis) height of the coil (cm).
        width: the horizontal (x axis) width of the coil (cm).
        center_point: the center of the coil (x,z) values (cm).
        name: defaults to "pf_coil".
    """

    def __init__(
        self,
        height: float,
        width: float,
        center_point: Tuple[float, float],
        name: Optional[str] = "pf_coil",
        **kwargs
    ) -> None:

        super().__init__(
            name=name,
            **kwargs
        )

        self.center_point = center_point
        self.height = height
        self.width = width

    @property
    def center_point(self):
        return self._center_point

    @center_point.setter
    def center_point(self, center_point):
        self._center_point = center_point

    @property
    def height(self):
        return self._height

    @height.setter
    def height(self, height):
        self._height = height

    @property
    def width(self):
        return self._width

    @width.setter
    def width(self, width):
        self._width = width

    def find_points(self):
        """Finds the XZ points joined by straight connections that describe
        the 2D profile of the poloidal field coil shape."""

        points = [
            (
                self.center_point[0] + self.width / 2.0,
                self.center_point[1] + self.height / 2.0,
            ),  # upper right
            (
                self.center_point[0] + self.width / 2.0,
                self.center_point[1] - self.height / 2.0,
            ),  # lower right
            (
                self.center_point[0] - self.width / 2.0,
                self.center_point[1] - self.height / 2.0,
            ),  # lower left
            (
                self.center_point[0] - self.width / 2.0,
                self.center_point[1] + self.height / 2.0,
            )
        ]

        self.points = points




from collections.abc import Iterable
from operator import itemgetter

from paramak import CuttingWedge

SAFETY_FACTOR = 3


class CuttingWedgeFS(CuttingWedge):
    """Creates a wedge from a Shape that can be useful for cutting sector
    models.

    Args:
        shape (paramak.Shape): a paramak.Shape object that is used to find the
            height and radius of the wedge
    """

    def __init__(
        self,
        shape,
        **kwargs
    ):
        self.shape = shape
        super().__init__(
            height=self.height,
            radius=self.radius,
            **kwargs
        )

    @property
    def shape(self):
        return self._shape

    @shape.setter
    def shape(self, value):
        if value.rotation_angle == 360:
            msg = ('cutting_wedge cannot be created, rotation_angle must be '
                   'less than 360 degrees')
            raise ValueError(msg)
        self._shape = value

    @property
    def radius(self):
        self.find_radius_height()
        return self._radius

    @radius.setter
    def radius(self, value):
        self._radius = value

    @property
    def height(self):
        self.find_radius_height()
        return self._height

    @height.setter
    def height(self, value):
        self._height = value

    @property
    def rotation_angle(self):
        self.rotation_angle = 360 - self.shape.rotation_angle
        return self._rotation_angle

    @rotation_angle.setter
    def rotation_angle(self, value):
        self._rotation_angle = value

    @property
    def workplane(self):
        workplanes = ["XY", "XZ", "YZ"]
        for wp in workplanes:
            if self.shape.get_rotation_axis()[1] in wp:
                self.workplane = wp
                break
        return self._workplane

    @workplane.setter
    def workplane(self, value):
        self._workplane = value

    @property
    def rotation_axis(self):
        self.rotation_axis = self.shape.rotation_axis
        return self._rotation_axis

    @rotation_axis.setter
    def rotation_axis(self, value):
        self._rotation_axis = value

    @property
    def azimuth_placement_angle(self):
        if isinstance(self.shape.azimuth_placement_angle, Iterable):
            self.azimuth_placement_angle = self.shape.rotation_angle
        else:
            self.azimuth_placement_angle = \
                self.shape.azimuth_placement_angle + self.shape.rotation_angle
        return self._azimuth_placement_angle

    @azimuth_placement_angle.setter
    def azimuth_placement_angle(self, value):
        self._azimuth_placement_angle = value

    def find_radius_height(self):
        shape = self.shape
        if shape.rotation_angle == 360:
            msg = ('cutting_wedge cannot be created, rotation_angle must be '
                   'less than 360 degrees')
            raise ValueError(msg)
        shape_points = shape.points
        if hasattr(shape, 'radius') and len(shape_points) == 1:
            max_x = shape_points[0][0] + shape.radius
            max_y = shape_points[0][1] + shape.radius

        elif len(shape_points) > 1:
            max_x = max(shape_points, key=itemgetter(0))[0]
            if shape.get_rotation_axis()[1] not in shape.workplane and \
                    hasattr(shape, "distance"):
                max_y = shape.distance
            else:
                max_y = max(shape_points, key=itemgetter(1))[1]

        else:
            raise ValueError('cutting_wedge cannot be created')
        self.radius = SAFETY_FACTOR * max_x
        self.height = SAFETY_FACTOR * max_y




from typing import Optional, Tuple

import cadquery as cq
from paramak import PoloidalFieldCoilSet, RotateStraightShape


class PoloidalFieldCoilCaseSetFC(RotateStraightShape):
    """Creates a series of rectangular poloidal field coils.

    Args:
        pf_coils (paramak.PoloidalFieldCoil): a list of pf coil objects or a
            CadQuery compound object
        casing_thicknesses (float or list of floats): the thicknesses of the
            coil casing (cm). If float then the same thickness is applied to
            all coils. If list of floats then each entry is applied to a
            seperate pf_coil, one entry for each pf_coil.
        name (str, optional): defaults to "pf_coil_case_set_fc".
    """

    def __init__(
        self,
        pf_coils,
        casing_thicknesses,
        name="pf_coil_case_set_fc",
        color: Optional[Tuple[int, int, int]] = (1., 1., 0.498),
        **kwargs
    ):

        super().__init__(
            name=name,
            color=color,
            **kwargs
        )

        self.casing_thicknesses = casing_thicknesses
        self.pf_coils = pf_coils

        # calculated internally by the class
        self.heights = None
        self.widths = None
        self.center_points = None

    @property
    def casing_thicknesses(self):
        return self._casing_thicknesses

    @casing_thicknesses.setter
    def casing_thicknesses(self, value):
        if isinstance(value, list):
            if not all(isinstance(x, (int, float)) for x in value):
                raise ValueError(
                    "Every entry in Casing_thicknesses must be a float or int")
        else:
            if not isinstance(value, (float, int)):
                raise ValueError(
                    "Casing_thicknesses must be a list of numbers or a number")
        self._casing_thicknesses = value

    @property
    def pf_coils(self):
        return self._pf_coils

    @pf_coils.setter
    def pf_coils(self, value):
        if not isinstance(value, (list, PoloidalFieldCoilSet)):
            msg = ('PoloidalFieldCoilCaseSetFC.pf_coils must be either a list '
                   'paramak.PoloidalFieldCoil or a '
                   'paramak.PoloidalFieldCoilSet object')
            raise ValueError(msg)
        self._pf_coils = value

    def find_points(self):
        """Finds the XZ points joined by straight connections that describe
        the 2D profile of the poloidal field coil shape."""

        if isinstance(self.pf_coils, list):
            self.heights = [entry.height for entry in self.pf_coils]
            self.widths = [entry.width for entry in self.pf_coils]
            self.center_points = [
                entry.center_point for entry in self.pf_coils]

            num_of_coils = len(self.pf_coils)

        elif isinstance(self.pf_coils, PoloidalFieldCoilSet):
            self.heights = self.pf_coils.heights
            self.widths = self.pf_coils.widths
            self.center_points = self.pf_coils.center_points
            num_of_coils = len(self.pf_coils.solid.Solids())

        if isinstance(self.casing_thicknesses, list):
            if len(self.casing_thicknesses) != num_of_coils:
                raise ValueError(
                    "The number pf_coils is not equal to the "
                    "number of thichnesses provided. "
                    "casing_thicknesses=",
                    self.casing_thicknesses,
                    "num_of_coils=",
                    num_of_coils)
            casing_thicknesses_list = self.casing_thicknesses
        else:
            casing_thicknesses_list = [self.casing_thicknesses] * num_of_coils

        all_points = []

        for height, width, center_point, casing_thickness in zip(
                self.heights, self.widths,
                self.center_points, casing_thicknesses_list):

            if casing_thickness != 0:

                all_points = all_points + [
                    (
                        center_point[0] + width / 2.0,
                        center_point[1] + height / 2.0,
                    ),  # upper right
                    (
                        center_point[0] + width / 2.0,
                        center_point[1] - height / 2.0,
                    ),  # lower right
                    (
                        center_point[0] - width / 2.0,
                        center_point[1] - height / 2.0,
                    ),  # lower left
                    (
                        center_point[0] - width / 2.0,
                        center_point[1] + height / 2.0,
                    ),  # upper left
                    (
                        center_point[0] + width / 2.0,
                        center_point[1] + height / 2.0,
                    ),  # upper right
                    (
                        center_point[0] + (casing_thickness + width / 2.0),
                        center_point[1] + (casing_thickness + height / 2.0),
                    ),
                    (
                        center_point[0] + (casing_thickness + width / 2.0),
                        center_point[1] - (casing_thickness + height / 2.0),
                    ),
                    (
                        center_point[0] - (casing_thickness + width / 2.0),
                        center_point[1] - (casing_thickness + height / 2.0),
                    ),
                    (
                        center_point[0] - (casing_thickness + width / 2.0),
                        center_point[1] + (casing_thickness + height / 2.0),
                    ),
                    (
                        center_point[0] + (casing_thickness + width / 2.0),
                        center_point[1] + (casing_thickness + height / 2.0),
                    )
                ]

        self.points = all_points

    def create_solid(self):
        """Creates a 3d solid using points with straight edges. Individual
        solids in the compound can be accessed using .Solids()[i] where i is an
        int

           Returns:
              A CadQuery solid: A 3D solid volume
        """

        iter_points = iter(self.points)
        pf_coils_set = []
        wires = []
        for p1, p2, p3, p4, p5, p6, p7, p8, p9, p10 in zip(
                iter_points, iter_points, iter_points, iter_points,
                iter_points, iter_points, iter_points, iter_points,
                iter_points, iter_points,
        ):

            solid = (
                cq.Workplane(self.workplane)
                .polyline(
                    [p1[:2], p2[:2], p3[:2], p4[:2], p5[:2], p6[:2],
                     p7[:2], p8[:2], p9[:2], p10[:2]])
            )

            wire = solid.close()

            wires.append(wire)

            solid = wire.revolve(self.rotation_angle)

            pf_coils_set.append(solid)

        compound = cq.Compound.makeCompound(
            [a.val() for a in pf_coils_set]
        )

        self.wire = wires

        self.solid = compound

        return compound




import unittest

import paramak


class TestDivertorITERNoDome(unittest.TestCase):

    def test_creation(self):
        """Creates an ITER-type divertor using the ITERtypeDivertorNoDome
        parametric component and checks that a cadquery solid is created."""

        test_shape = paramak.ITERtypeDivertorNoDome()
        assert test_shape.solid is not None

    def test_stp_export(self):
        """Creates an ITER-type divertor using the ITERtypeDivertorNoDome
        parametric component and checks that a stp file of the shape can be
        exported using the export_stp method."""

        test_shape = paramak.ITERtypeDivertorNoDome()
        test_shape.export_stp("tests/ITER_div_no_dome.step")

    def test_faces(self):
        """Creates an ITER-type divertor using the ITERtypeDivertorNoDome
        parametric component and checks that a solid with the correct number
        of faces is created"""

        test_shape = paramak.ITERtypeDivertorNoDome()
        assert len(test_shape.areas) == 10
        assert len(set(test_shape.areas)) == 10

        test_shape.rotation_angle = 180
        assert len(test_shape.areas) == 12
        assert len(set(test_shape.areas)) == 11




import paramak


class SingleNullSubmersionTokamak(paramak.SubmersionTokamak):
    """Creates geometry for a submersion reactor with a single divertor
    including a plasma, cylindrical center column shielding, square toroidal
    field coils. There is an inboard breeder blanket on this submersion
    reactor.

    Arguments:
        divertor_position (str): Defaults to "upper".
        support_position (str): Defaults to "upper".
    """

    def __init__(
        self,
        divertor_position: str = "upper",
        support_position: str = "upper",
        **kwargs
    ):

        super().__init__(
            divertor_position=divertor_position,
            support_position=support_position,
            **kwargs)




import warnings
from typing import Callable, List, Optional, Union

import mpmath
import numpy as np
import paramak
import sympy as sp
from paramak import RotateMixedShape, diff_between_angles
from paramak.parametric_components.tokamak_plasma_plasmaboundaries import \
    PlasmaBoundaries
from scipy.interpolate import interp1d


class BlanketFP(RotateMixedShape):
    """A blanket volume created from plasma parameters.

    Args:
        thickness (float or [float] or callable or [(float), (float)]):
            the thickness of the blanket (cm). If the thickness is a float then
            this produces a blanket of constant thickness. If the thickness is
            a tuple of floats, blanket thickness will vary linearly between the
            two values. If thickness is callable, then the blanket thickness
            will be a function of poloidal angle (in degrees). If thickness is
            a list of two lists (thicknesses and angles) then these will be
            used together with linear interpolation.
        start_angle: the angle in degrees to start the blanket, measured anti
            clockwise from 3 o'clock.
        stop_angle: the angle in degrees to stop the blanket, measured anti
            clockwise from 3 o'clock.
        plasma: If not None, the parameters of the plasma Object will be used.
        minor_radius: the minor radius of the plasma (cm).
        major_radius: the major radius of the plasma (cm).
        triangularity: the triangularity of the plasma.
        elongation: the elongation of the plasma.
        vertical_displacement: the vertical_displacement of the plasma (cm).
        offset_from_plasma: the distance between the plasma and the blanket
            (cm). If float, constant offset. If list of floats, offset will
            vary linearly between the values. If callable, offset will be a
            function of poloidal angle (in degrees). If a list of two lists
            (angles and offsets) then these will be used together with linear
            interpolation.
        num_points: number of points that will describe the shape.
    """

    def __init__(self,
                 thickness,
                 start_angle: float,
                 stop_angle: float,
                 plasma: Optional[Union[paramak.Plasma,
                                        paramak.PlasmaBoundaries,
                                        paramak.PlasmaFromPoints]] = None,
                 minor_radius: Optional[float] = 150.0,
                 major_radius: Optional[float] = 450.0,
                 triangularity: Optional[float] = 0.55,
                 elongation: Optional[float] = 2.0,
                 vertical_displacement: Optional[float] = 0.0,
                 offset_from_plasma: Optional[float] = 0.0,
                 num_points: Optional[int] = 50,
                 **kwargs):

        super().__init__(**kwargs)

        self.thickness = thickness
        self.start_angle, self.stop_angle = None, None
        self.start_angle = start_angle
        self.stop_angle = stop_angle
        self.plasma = plasma
        self.vertical_displacement = vertical_displacement
        if plasma is None:
            self.minor_radius = minor_radius
            self.major_radius = major_radius
            self.triangularity = triangularity
            self.elongation = elongation
        else:  # if plasma object is given, use its parameters
            self.minor_radius = plasma.minor_radius
            self.major_radius = plasma.major_radius
            self.triangularity = plasma.triangularity
            self.elongation = plasma.elongation
        self.offset_from_plasma = offset_from_plasma
        self.num_points = num_points
        self.physical_groups = None

    @property
    def start_angle(self):
        return self._start_angle

    @start_angle.setter
    def start_angle(self, value):
        self._start_angle = value

    @property
    def stop_angle(self):
        return self._stop_angle

    @stop_angle.setter
    def stop_angle(self, value):
        self._stop_angle = value

    @property
    def physical_groups(self):
        self.create_physical_groups()
        return self._physical_groups

    @physical_groups.setter
    def physical_groups(self, physical_groups):
        self._physical_groups = physical_groups

    @property
    def minor_radius(self):
        return self._minor_radius

    @minor_radius.setter
    def minor_radius(self, minor_radius):
        self._minor_radius = minor_radius

    @property
    def thickness(self):
        return self._thickness

    @thickness.setter
    def thickness(self, thickness):
        self._thickness = thickness

    @property
    def inner_points(self):
        self.find_points()
        return self._inner_points

    @inner_points.setter
    def inner_points(self, value):
        self._inner_points = value

    @property
    def outer_points(self):
        self.find_points()
        return self._outer_points

    @outer_points.setter
    def outer_points(self, value):
        self._outer_points = value

    def make_callable(self, attribute):
        """This function transforms an attribute (thickness or offset) into a
        callable function of theta
        """
        # if the attribute is a list, create a interpolated object of the
        # values
        if isinstance(attribute, (tuple, list)):
            if isinstance(attribute[0], (tuple, list)) and \
                isinstance(attribute[1], (tuple, list)) and \
                    len(attribute) == 2:
                # attribute is a list of 2 lists
                if len(attribute[0]) != len(attribute[1]):
                    raise ValueError('The length of angles list must equal \
                     the length of values list')
                list_of_angles = np.array(attribute[0])
                offset_values = attribute[1]
            else:
                # no list of angles is given
                offset_values = attribute
                list_of_angles = np.linspace(
                    self.start_angle,
                    self.stop_angle,
                    len(offset_values),
                    endpoint=True)
            interpolated_values = interp1d(list_of_angles, offset_values)

        def fun(theta):
            if callable(attribute):
                return attribute(theta)
            elif isinstance(attribute, (tuple, list)):
                return interpolated_values(theta)
            else:
                return attribute
        return fun

    def find_points(self, angles=None):
        self._overlapping_shape = False
        # create array of angles theta
        if angles is None:
            thetas = np.linspace(
                self.start_angle,
                self.stop_angle,
                num=self.num_points,
                endpoint=True,
            )
        else:
            thetas = angles

        # create inner points
        inner_offset = self.make_callable(self.offset_from_plasma)
        inner_points = self.create_offset_points(thetas, inner_offset)
        inner_points[-1][2] = "straight"
        self.inner_points = inner_points

        # create outer points
        thickness = self.make_callable(self.thickness)

        def outer_offset(theta):
            return inner_offset(theta) + thickness(theta)

        outer_points = self.create_offset_points(np.flip(thetas), outer_offset)
        outer_points[-1][2] = "straight"
        self.outer_points = outer_points

        # assemble
        points = inner_points + outer_points
        if self._overlapping_shape:
            msg = ("BlanketFP: Some points with negative R coordinate have "
                   "been ignored.")
            warnings.warn(msg)

        self.points = points
        return points

    def create_offset_points(self, thetas, offset):
        """generates a list of points following parametric equations with an
        offset

        Args:
            thetas (np.array): the angles in degrees.
            offset (callable): offset value (cm). offset=0 will follow the
                parametric equations.

        Returns:
            list: list of points [[R1, Z1, connection1], [R2, Z2, connection2],
            ...]
        """
        # create sympy objects and derivatives
        theta_sp = sp.Symbol("theta")

        R_sp, Z_sp = self.distribution(theta_sp, pkg=sp)
        R_derivative = sp.diff(R_sp, theta_sp)
        Z_derivative = sp.diff(Z_sp, theta_sp)
        points = []

        for theta in thetas:
            # get local value of derivatives
            val_R_derivative = float(R_derivative.subs("theta", theta))
            val_Z_derivative = float(Z_derivative.subs("theta", theta))

            # get normal vector components
            nx = val_Z_derivative
            ny = -val_R_derivative

            # normalise normal vector
            normal_vector_norm = (nx ** 2 + ny ** 2) ** 0.5
            nx /= normal_vector_norm
            ny /= normal_vector_norm

            # calculate outer points
            val_R_outer = self.distribution(theta)[0] + offset(theta) * nx
            val_Z_outer = self.distribution(theta)[1] + offset(theta) * ny
            if float(val_R_outer) > 0:
                points.append(
                    [float(val_R_outer), float(val_Z_outer), "spline"])
            else:
                self._overlapping_shape = True
        return points

    def create_physical_groups(self):
        """Creates the physical groups for STP files

        Returns:
            list: list of dicts containing the physical groups
        """

        groups = []
        number_of_volumes = 1  # only one volume
        number_of_surfaces = 2  # inner and outer

        surface_names = ["inner", "outer"]
        volumes_names = ["inside"]

        # add two cut sections if they exist
        if self.rotation_angle != 360:
            number_of_surfaces += 2
            surface_names += ["left_section", "right_section"]
            full_rotation = False
        else:
            full_rotation = True

        # add two surfaces between blanket and div if they exist
        if diff_between_angles(self.start_angle, self.stop_angle) != 0:
            number_of_surfaces += 2
            surface_names += ["inner_section", "outer_section"]
            stop_equals_start = False
        else:
            stop_equals_start = True

        # rearrange order
        new_order = [i for i in range(len(surface_names))]
        if full_rotation:
            if not stop_equals_start:
                # from ["inner", "outer", "inner_section", "outer_section"]

                # to ["inner", "inner_section", "outer", "outer_section"]
                new_order = [0, 2, 1, 3]
        else:
            if stop_equals_start:
                print(
                    "Warning: If start_angle = stop_angle surfaces will not\
                     be handled correctly"
                )
                new_order = [0, 1, 2, 3]
            else:
                # from ['inner', 'outer', 'left_section', 'right_section',
                #           'inner_section', 'outer_section']

                # to ["inner", "inner_section", "outer", "outer_section",
                #         "left_section", "right_section"]
                new_order = [0, 4, 1, 5, 2, 3]
        surface_names = [surface_names[i] for i in new_order]

        for i in range(1, number_of_volumes + 1):
            group = {"dim": 3, "id": i, "name": volumes_names[i - 1]}
            groups.append(group)
        for i in range(1, number_of_surfaces + 1):
            group = {"dim": 2, "id": i, "name": surface_names[i - 1]}
            groups.append(group)
        self.physical_groups = groups

    def distribution(self, theta, pkg=np):
        """Plasma distribution theta in degrees

        Args:
            theta (float or np.array or sp.Symbol): the angle(s) in degrees.
            pkg (module, optional): Module to use in the funciton. If sp, as
                sympy object will be returned. If np, a np.array or a float
                will be returned. Defaults to np.

        Returns:
            (float, float) or (sympy.Add, sympy.Mul) or
                (numpy.array, numpy.array): The R and Z coordinates of the
                point with angle theta
        """
        if pkg == np:
            theta = np.radians(theta)
        else:
            theta = mpmath.radians(theta)
        R = self.major_radius + self.minor_radius * pkg.cos(
            theta + self.triangularity * pkg.sin(theta)
        )
        Z = (
            self.elongation * self.minor_radius * pkg.sin(theta)
            + self.vertical_displacement
        )
        return R, Z




from typing import Optional, Tuple

from paramak import Shape


class RotateMixedShape(Shape):
    """Rotates a 3d CadQuery solid from points connected with a mixture of
    straight lines and splines.

    Args:
        rotation_angle: The rotation_angle to use when revolving the solid
            (degrees). Defaults to 360.0.
    """

    def __init__(
        self,
        rotation_angle: Optional[float] = 360.,
        color: Optional[Tuple[float, float, float, Optional[float]]] = (0.121, 0.47, 0.705),
        **kwargs
    ):

        super().__init__(
            color=color,
            **kwargs
        )
        self.rotation_angle = rotation_angle

    @property
    def rotation_angle(self):
        return self._rotation_angle

    @rotation_angle.setter
    def rotation_angle(self, value):
        self._rotation_angle = value

    def create_solid(self):
        """Creates a rotated 3d solid using points with straight and spline
        edges.

           Returns:
              A CadQuery solid: A 3D solid volume
        """

        solid = super().create_solid()

        wire = solid.close()

        self.wire = wire

        solid = wire.revolve(self.rotation_angle)

        solid = self.rotate_solid(solid)
        solid = self.perform_boolean_operations(solid)
        self.solid = solid
        return solid




import os
import unittest
from pathlib import Path

import pytest
from paramak import ExtrudeSplineShape


class TestExtrudeSplineShape(unittest.TestCase):

    def setUp(self):
        self.test_shape = ExtrudeSplineShape(
            points=[(50, 0), (50, 20), (70, 80), (90, 50), (70, 0), (90, -50),
                    (70, -80), (50, -20)], distance=30
        )

    def test_default_parameters(self):
        """Checks that the default parameters of an ExtrudeSplineShape are
        correct."""

        assert self.test_shape.rotation_angle == 360
        assert self.test_shape.extrude_both

    def test_absolute_shape_volume(self):
        """Creates an ExtrudeSplineShape and checks that the volume is
        correct."""

        assert self.test_shape.solid is not None
        assert self.test_shape.volume() > 20 * 20 * 30

    def test_shape_face_areas(self):
        """Creates an ExtrudeSplineShape and checks that the face areas are
        expected."""

        self.test_shape.extrude_both = False
        assert len(self.test_shape.areas) == 3
        assert len(set([round(i) for i in self.test_shape.areas])) == 2

    def test_relative_shape_volume(self):
        """Creates two ExtrudeSplineShapes and checks that their relative
        volumes are correct."""

        test_volume = self.test_shape.volume()
        self.test_shape.azimuth_placement_angle = [0, 180]

        assert self.test_shape.volume() == pytest.approx(
            test_volume * 2, rel=0.01)

    def test_cut_volume(self):
        """Creates an ExtrudeSplineShape with another ExtrudeSplineShape cut out
        and checks that the volume is correct."""

        inner_shape = ExtrudeSplineShape(
            points=[(5, 5), (5, 10), (10, 10), (10, 5)], distance=30
        )

        outer_shape = ExtrudeSplineShape(
            points=[(3, 3), (3, 12), (12, 12), (12, 3)], distance=30
        )

        outer_shape_with_cut = ExtrudeSplineShape(
            points=[(3, 3), (3, 12), (12, 12), (12, 3)], cut=inner_shape,
            distance=30,
        )

        assert inner_shape.volume() == pytest.approx(1165, abs=2)
        assert outer_shape.volume() == pytest.approx(3775, abs=2)
        assert outer_shape_with_cut.volume() == pytest.approx(3775 - 1165, abs=2)

    def test_rotation_angle(self):
        """Creates an ExtrudeStraightShape with a rotation_angle < 360 and
        checks that the correct cut is performed and the volume is correct."""

        self.test_shape.azimuth_placement_angle = [45, 135, 225, 315]
        test_volume = self.test_shape.volume()
        self.test_shape.rotation_angle = 180
        assert self.test_shape.volume() == pytest.approx(
            test_volume * 0.5, rel=0.01)

    def test_extrude_both(self):
        """Creates an ExtrudeSplineShape with extrude_both = True and False and
        checks that the volumes are correct."""

        test_volume_extrude_both = self.test_shape.volume()
        self.test_shape.extrude_both = False
        assert self.test_shape.volume() == pytest.approx(
            test_volume_extrude_both)

    def test_export_stp(self):
        """Exports and stp file with mode = solid and wire and checks that the
        outputs exist and relative file sizes are correct."""

        os.system("rm test_solid.stp test_solid2.stp test_wire.stp")

        self.test_shape.export_stp('test_solid.stp', mode='solid')
        self.test_shape.export_stp('test_solid2.stp')
        self.test_shape.export_stp('test_wire.stp', mode='wire')

        assert Path("test_solid.stp").exists() is True
        assert Path("test_solid2.stp").exists() is True
        assert Path("test_wire.stp").exists() is True

        assert Path("test_solid.stp").stat().st_size == \
            Path("test_solid2.stp").stat().st_size
        assert Path("test_wire.stp").stat().st_size < \
            Path("test_solid2.stp").stat().st_size

        os.system("rm test_solid.stp test_solid2.stp test_wire.stp")

    def test_incorrect_points_input(self):
        """Checks that an error is raised when the points are input with the
        connection"""

        def incorrect_points_definition():
            self.test_shape.points = [
                (10, 10, 'spline'),
                (10, 30, 'spline'),
                (30, 30, 'spline'),
                (30, 10, 'spline')
            ]

        self.assertRaises(
            ValueError,
            incorrect_points_definition
        )


if __name__ == "__main__":
    unittest.main()




import warnings
from typing import List, Optional, Union

import paramak


class BallReactor(paramak.Reactor):
    """Creates geometry for a simple ball reactor including a plasma,
    cylindrical center column shielding, square toroidal field coils.
    There is no inboard breeder blanket on this ball reactor like
    most spherical reactors.

    Arguments:
        inner_bore_radial_thickness: the radial thickness of the inner bore
            (cm)
        inboard_tf_leg_radial_thickness: the radial thickness of the inner leg
            of the toroidal field coils (cm)
        center_column_shield_radial_thickness: the radial thickness of the
            center column shield (cm)
        divertor_radial_thickness: the radial thickness of the divertor
            (cm), this fills the gap between the center column shield and
            blanket
        inner_plasma_gap_radial_thickness: the radial thickness of the
            inboard gap between the plasma and the center column shield (cm)
        plasma_radial_thickness: the radial thickness of the plasma
        outer_plasma_gap_radial_thickness: the radial thickness of the
            outboard gap between the plasma and firstwall (cm)
        firstwall_radial_thickness: the radial thickness of the first wall (cm)
        blanket_radial_thickness: the radial thickness of the blanket (cm)
        blanket_rear_wall_radial_thickness: the radial thickness of the rear
            wall of the blanket (cm)
        elongation: the elongation of the plasma
        triangularity: the triangularity of the plasma
        plasma_gap_vertical_thickness: the vertical thickness of the gap
            between the plasma and firstwall (cm).
        divertor_to_tf_gap_vertical_thickness: the vertical thickness of the
            gap between the divertor and the TF coils.
        number_of_tf_coils: the number of tf coils
        pf_coil_radial_thicknesses: the radial
            thickness of each poloidal field coil.
        pf_coil_vertical_thicknesses: the vertical
            thickness of each poloidal field coil.
        pf_coil_to_tf_coil_radial_gap: the radial distance
            between the rear of the poloidal field coil and the toroidal field
            coil.
        pf_coil_radial_position: The radial (x) position(s) of the centers of
            the poloidal field coils.
        pf_coil_vertical_position: The vertical (z) position(s) of the centers
            of the poloidal field coils.
        pf_coil_case_thicknesses: the thickness(s) to use in both the radial
            and vertical direction for the casing around the pf coils. If float
            then the single value will be applied to all pf coils. If list then
            each value will be applied to the pf coils one by one. To have no
            casing set to 0.
        outboard_tf_coil_radial_thickness: the radial thickness of the toroidal
            field coil.
        outboard_tf_coil_poloidal_thickness: the poloidal thickness of the
            toroidal field coil.
        divertor_position: the position of the divertor, "upper", "lower" or
            "both".
        rotation_angle: the angle of the sector that is desired.
    """

    def __init__(
            self,
            inner_bore_radial_thickness: float,
            inboard_tf_leg_radial_thickness: float,
            center_column_shield_radial_thickness: float,
            divertor_radial_thickness: float,
            inner_plasma_gap_radial_thickness: float,
            plasma_radial_thickness: float,
            outer_plasma_gap_radial_thickness: float,
            firstwall_radial_thickness: float,
            blanket_radial_thickness: float,
            blanket_rear_wall_radial_thickness: float,
            elongation: float,
            triangularity: float,
            plasma_gap_vertical_thickness: float,
            divertor_to_tf_gap_vertical_thickness: Optional[float] = 0,
            number_of_tf_coils: Optional[int] = 12,
            rear_blanket_to_tf_gap: Optional[float] = None,
            pf_coil_radial_thicknesses: Optional[Union[float, List[float]]] = None,
            pf_coil_vertical_thicknesses: Optional[Union[float, List[float]]] = None,
            pf_coil_radial_position: Optional[Union[float, List[float]]] = None,
            pf_coil_vertical_position: Optional[Union[float, List[float]]] = None,
            pf_coil_case_thicknesses: Optional[Union[float, List[float]]] = 10,
            outboard_tf_coil_radial_thickness: float = None,
            outboard_tf_coil_poloidal_thickness: float = None,
            divertor_position: Optional[str] = "both",
            rotation_angle: Optional[str] = 360.0,
    ):

        super().__init__([])

        self.inner_bore_radial_thickness = inner_bore_radial_thickness
        self.inboard_tf_leg_radial_thickness = inboard_tf_leg_radial_thickness
        self.center_column_shield_radial_thickness = \
            center_column_shield_radial_thickness
        self.divertor_radial_thickness = divertor_radial_thickness
        self.inner_plasma_gap_radial_thickness = \
            inner_plasma_gap_radial_thickness
        self.plasma_radial_thickness = plasma_radial_thickness
        self.outer_plasma_gap_radial_thickness = \
            outer_plasma_gap_radial_thickness
        self.firstwall_radial_thickness = firstwall_radial_thickness
        self.blanket_radial_thickness = blanket_radial_thickness
        self.blanket_rear_wall_radial_thickness = \
            blanket_rear_wall_radial_thickness

        self.pf_coil_vertical_position = pf_coil_vertical_position
        self.pf_coil_radial_position = pf_coil_radial_position
        self.pf_coil_radial_thicknesses = pf_coil_radial_thicknesses
        self.pf_coil_vertical_thicknesses = pf_coil_vertical_thicknesses
        self.rear_blanket_to_tf_gap = rear_blanket_to_tf_gap

        self.pf_coil_case_thicknesses = pf_coil_case_thicknesses
        self.outboard_tf_coil_radial_thickness = \
            outboard_tf_coil_radial_thickness
        self.outboard_tf_coil_poloidal_thickness = \
            outboard_tf_coil_poloidal_thickness
        self.divertor_position = divertor_position
        self.rotation_angle = rotation_angle

        self.plasma_gap_vertical_thickness = plasma_gap_vertical_thickness
        self.divertor_to_tf_gap_vertical_thickness = divertor_to_tf_gap_vertical_thickness

        self.elongation = elongation
        self.triangularity = triangularity

        self.number_of_tf_coils = number_of_tf_coils
        self.rotation_angle = rotation_angle

        # adds self.input_variable_names from the Reactor class
        self.input_variable_names = self.input_variable_names + [
            'inner_bore_radial_thickness',
            'inboard_tf_leg_radial_thickness',
            'center_column_shield_radial_thickness',
            'divertor_radial_thickness',
            'inner_plasma_gap_radial_thickness',
            'plasma_radial_thickness',
            'outer_plasma_gap_radial_thickness',
            'firstwall_radial_thickness',
            'blanket_radial_thickness',
            'blanket_rear_wall_radial_thickness',
            'elongation',
            'triangularity',
            'plasma_gap_vertical_thickness',
            'divertor_to_tf_gap_vertical_thickness',
            'number_of_tf_coils',
            'rear_blanket_to_tf_gap',
            'pf_coil_radial_thicknesses',
            'pf_coil_vertical_thicknesses',
            'pf_coil_radial_position',
            'pf_coil_vertical_position',
            'pf_coil_case_thicknesses',
            'outboard_tf_coil_radial_thickness',
            'outboard_tf_coil_poloidal_thickness',
            'divertor_position',
            'rotation_angle',
        ]

        # set by make_plasma
        self.major_radius = None
        self.minor_radius = None

    @property
    def rotation_angle(self):
        return self._rotation_angle

    @rotation_angle.setter
    def rotation_angle(self, value):
        if value == 360:
            msg = (
                "360 degree rotation may result in a "
                "Standard_ConstructionError or AttributeError"
            )
            warnings.warn(msg, UserWarning)
        elif value > 360:
            raise ValueError('rotation_angle can not be larger than 360')
        self._rotation_angle = value

    @property
    def pf_coil_vertical_position(self):
        return self._pf_coil_vertical_position

    @pf_coil_vertical_position.setter
    def pf_coil_vertical_position(self, value):
        if not isinstance(value, list) and value is not None:
            raise ValueError("pf_coil_vertical_position must be a list")
        self._pf_coil_vertical_position = value

    @property
    def pf_coil_radial_position(self):
        return self._pf_coil_radial_position

    @pf_coil_radial_position.setter
    def pf_coil_radial_position(self, value):
        if not isinstance(value, list) and value is not None:
            raise ValueError("pf_coil_radial_position must be a list")
        self._pf_coil_radial_position = value

    @property
    def pf_coil_radial_thicknesses(self):
        return self._pf_coil_radial_thicknesses

    @pf_coil_radial_thicknesses.setter
    def pf_coil_radial_thicknesses(self, value):
        if not isinstance(value, list) and value is not None:
            raise ValueError("pf_coil_radial_thicknesses must be a list")
        self._pf_coil_radial_thicknesses = value

    @property
    def pf_coil_vertical_thicknesses(self):
        return self._pf_coil_vertical_thicknesses

    @pf_coil_vertical_thicknesses.setter
    def pf_coil_vertical_thicknesses(self, value):
        if not isinstance(value, list) and value is not None:
            raise ValueError("pf_coil_vertical_thicknesses must be a list")
        self._pf_coil_vertical_thicknesses = value

    @property
    def divertor_position(self):
        return self._divertor_position

    @divertor_position.setter
    def divertor_position(self, value):
        acceptable_values = ["upper", "lower", "both"]
        if value in acceptable_values:
            self._divertor_position = value
        else:
            msg = "divertor_position must be 'upper', 'lower' or 'both'"
            raise ValueError(msg)

    def create_solids(self):
        """Creates a list of paramak.Shape for components and saves it in
        self.shapes_and_components
        """

        uncut_shapes = []

        uncut_shapes.append(self._make_plasma())
        self._make_radial_build()
        self._make_vertical_build()
        uncut_shapes.append(self._make_inboard_tf_coils())
        uncut_shapes.append(self._make_center_column_shield())
        uncut_shapes += self._make_blankets_layers()
        uncut_shapes.append(self._make_divertor())
        uncut_shapes += self._make_tf_coils()
        pf_coils = self._make_pf_coils()

        if pf_coils is None:
            shapes_and_components = uncut_shapes
        else:
            for shape in uncut_shapes:
                for pf_coil in pf_coils:
                    shape.solid = shape.solid.cut(pf_coil.solid)
            shapes_and_components = pf_coils + uncut_shapes

        self.shapes_and_components = shapes_and_components

    def _make_plasma(self):

        inner_equatorial_point = (
            self.inner_bore_radial_thickness
            + self.inboard_tf_leg_radial_thickness
            + self.center_column_shield_radial_thickness
            + self.inner_plasma_gap_radial_thickness
        )
        outer_equatorial_point = inner_equatorial_point + self.plasma_radial_thickness

        # sets major radius and minor radius from equatorial_points to allow a
        # radial build. This helps avoid the plasma overlapping the center
        # column and other components
        self.major_radius = (outer_equatorial_point +
                             inner_equatorial_point) / 2
        self.minor_radius = self.major_radius - inner_equatorial_point

        plasma = paramak.Plasma(
            major_radius=self.major_radius,
            minor_radius=self.minor_radius,
            elongation=self.elongation,
            triangularity=self.triangularity,
            rotation_angle=self.rotation_angle,
        )

        self._plasma = plasma
        return self._plasma

    def _make_radial_build(self):

        # this is the radial build sequence, where one component stops and
        # another starts

        self._inner_bore_start_radius = 0
        self._inner_bore_end_radius = (
            self._inner_bore_start_radius + self.inner_bore_radial_thickness
        )

        self._inboard_tf_coils_start_radius = self._inner_bore_end_radius
        self._inboard_tf_coils_end_radius = (
            self._inboard_tf_coils_start_radius +
            self.inboard_tf_leg_radial_thickness)

        self._center_column_shield_start_radius = \
            self._inboard_tf_coils_end_radius
        self._center_column_shield_end_radius = (
            self._center_column_shield_start_radius
            + self.center_column_shield_radial_thickness
        )

        self._divertor_start_radius = self._center_column_shield_end_radius
        self._divertor_end_radius = (
            self._center_column_shield_end_radius +
            self.divertor_radial_thickness)

        self._firstwall_start_radius = (
            self._center_column_shield_end_radius
            + self.inner_plasma_gap_radial_thickness
            + self.plasma_radial_thickness
            + self.outer_plasma_gap_radial_thickness
        )
        self._firstwall_end_radius = self._firstwall_start_radius + \
            self.firstwall_radial_thickness

        self._blanket_start_radius = self._firstwall_end_radius
        self._blanket_end_radius = \
            self._blanket_start_radius + self.blanket_radial_thickness

        self._blanket_rear_wall_start_radius = self._blanket_end_radius
        self._blanket_rear_wall_end_radius = (
            self._blanket_rear_wall_start_radius +
            self.blanket_rear_wall_radial_thickness)

    def _make_vertical_build(self):

        # this is the vertical build sequence, components build on each other
        # in a similar manner to the radial build

        self._firstwall_start_height = (
            self._plasma.high_point[1] +
            self.plasma_gap_vertical_thickness)
        self._firstwall_end_height = self._firstwall_start_height + \
            self.firstwall_radial_thickness

        self._blanket_start_height = self._firstwall_end_height
        self._blanket_end_height = self._blanket_start_height + self.blanket_radial_thickness

        self._blanket_rear_wall_start_height = self._blanket_end_height
        self._blanket_rear_wall_end_height = self._blanket_rear_wall_start_height + \
            self.blanket_rear_wall_radial_thickness

        self._tf_coil_start_height = self._blanket_rear_wall_end_height + \
            self.divertor_to_tf_gap_vertical_thickness

        self._center_column_shield_height = self._blanket_rear_wall_end_height * 2

        if self.rear_blanket_to_tf_gap is not None:
            self._tf_coil_start_radius = self._blanket_rear_wall_end_radius + \
                self.rear_blanket_to_tf_gap
            self._tf_coil_end_radius = (
                self._tf_coil_start_radius +
                self.outboard_tf_coil_radial_thickness)

    def _make_inboard_tf_coils(self):

        self._inboard_tf_coils = paramak.CenterColumnShieldCylinder(
            height=self._tf_coil_start_height * 2,
            inner_radius=self._inboard_tf_coils_start_radius,
            outer_radius=self._inboard_tf_coils_end_radius,
            rotation_angle=self.rotation_angle,
            name="inboard_tf_coils",
            color=(0, 0, 1)
        )
        return self._inboard_tf_coils

    def _make_center_column_shield(self):

        self._center_column_shield = paramak.CenterColumnShieldCylinder(
            height=self._center_column_shield_height,
            inner_radius=self._center_column_shield_start_radius,
            outer_radius=self._center_column_shield_end_radius,
            rotation_angle=self.rotation_angle,
            color=(0., 0.333, 0.),
            name="center_column_shield",
        )
        return self._center_column_shield

    def _make_blankets_layers(self):

        offset_from_plasma = [
            self.major_radius - self.minor_radius,
            self.plasma_gap_vertical_thickness,
            self.outer_plasma_gap_radial_thickness,
            self.plasma_gap_vertical_thickness,
            self.major_radius - self.minor_radius
        ]

        self._center_column_cutter = paramak.CenterColumnShieldCylinder(
            # extra 0.5 to ensure overlap,
            height=self._center_column_shield_height * 1.5,
            inner_radius=0,
            outer_radius=self._center_column_shield_end_radius,
            rotation_angle=360,
            color=(0., 0., 1.)
        )

        self._firstwall = paramak.BlanketFP(
            plasma=self._plasma,
            thickness=self.firstwall_radial_thickness,
            offset_from_plasma=offset_from_plasma,
            start_angle=-180,
            stop_angle=180,
            rotation_angle=self.rotation_angle,
            color=(0.5, 0.5, 0.5),
            name='firstwall',
            cut=[self._center_column_cutter]
        )

        self._blanket = paramak.BlanketFP(
            plasma=self._plasma,
            thickness=self.blanket_radial_thickness,
            offset_from_plasma=[e + self.firstwall_radial_thickness
                                for e in offset_from_plasma],
            start_angle=-180,
            stop_angle=180,
            rotation_angle=self.rotation_angle,
            color=(0., 1., 0.498),
            name='blanket',
            cut=[self._center_column_cutter])

        self._blanket_rear_wall = paramak.BlanketFP(
            plasma=self._plasma,
            thickness=self.blanket_rear_wall_radial_thickness,
            offset_from_plasma=[e + self.firstwall_radial_thickness
                                + self.blanket_radial_thickness
                                for e in offset_from_plasma],
            start_angle=-180,
            stop_angle=180,
            rotation_angle=self.rotation_angle,
            color=(0., 1., 1.),
            name='blanket_rear_wall',
            cut=[self._center_column_cutter],
        )

        return [self._firstwall, self._blanket, self._blanket_rear_wall]

    def _make_divertor(self):

        offset_from_plasma = [
            self.major_radius - self.minor_radius,
            self.plasma_gap_vertical_thickness,
            self.outer_plasma_gap_radial_thickness,
            self.plasma_gap_vertical_thickness,
            self.major_radius - self.minor_radius
        ]

        # used as an intersect when making the divertor
        self._blanket_fw_rear_wall_envelope = paramak.BlanketFP(
            plasma=self._plasma,
            thickness=self.firstwall_radial_thickness +
            self.blanket_radial_thickness +
            self.blanket_rear_wall_radial_thickness,
            offset_from_plasma=offset_from_plasma,
            start_angle=-180,
            stop_angle=180,
            rotation_angle=self.rotation_angle,
        )

        divertor_height = self._blanket_rear_wall_end_height * 2

        divertor_height_top = divertor_height
        divertor_height_bottom = -divertor_height

        if self.divertor_position == "lower":
            divertor_height_top = 0
        elif self.divertor_position == "upper":
            divertor_height_bottom = 0
        self._divertor = paramak.RotateStraightShape(
            points=[
                (self._divertor_start_radius, divertor_height_bottom),
                (self._divertor_end_radius, divertor_height_bottom),
                (self._divertor_end_radius, divertor_height_top),
                (self._divertor_start_radius, divertor_height_top)
            ],
            intersect=self._blanket_fw_rear_wall_envelope,
            name="divertor",
            color=(1., 0.667, 0.),
            rotation_angle=self.rotation_angle
        )

        for component in [
                self._firstwall,
                self._blanket,
                self._blanket_rear_wall]:
            component.cut.append(self._divertor)

        return self._divertor

    def _make_pf_coils(self):
        if None not in [self.pf_coil_vertical_thicknesses,
                        self.pf_coil_radial_thicknesses,
                        self.pf_coil_vertical_position,
                        self.pf_coil_radial_position]:
            list_of_components = []

            # TODO make use of counter in the name attribute

            center_points = [
                (x, y) for x, y in zip(
                    self.pf_coil_radial_position, self.pf_coil_vertical_position)]

            self._pf_coils = self._pf_coil = paramak.PoloidalFieldCoilSet(
                heights=self.pf_coil_vertical_thicknesses,
                widths=self.pf_coil_radial_thicknesses,
                center_points=center_points,
                rotation_angle=self.rotation_angle,
                name="pf_coil",
            )
            list_of_components.append(self._pf_coil)

            if self.pf_coil_case_thicknesses is not None:
                self._pf_coils_casing = paramak.PoloidalFieldCoilCaseSetFC(
                    pf_coils=self._pf_coil,
                    casing_thicknesses=self.pf_coil_case_thicknesses,
                    rotation_angle=self.rotation_angle,
                    name="pf_coil_case",
                )
                list_of_components.append(self._pf_coils_casing)

            return list_of_components
        else:
            print(
                'pf_coil_vertical_thicknesses, pf_coil_radial_thicknesses, '
                'pf_coil_radial_position, pf_coil_vertical_position not '
                'so not making pf coils')
            return None

    def _make_tf_coils(self):
        comp = []
        # checks that all the required information has been input by the user
        if None not in [
                self.rear_blanket_to_tf_gap,
                self.outboard_tf_coil_radial_thickness,
                self.outboard_tf_coil_poloidal_thickness,
                self.number_of_tf_coils] and self.number_of_tf_coils > 1:

            self._tf_coil = paramak.ToroidalFieldCoilRectangle(
                with_inner_leg=False,
                horizontal_start_point=(
                    self._inboard_tf_coils_start_radius,
                    self._tf_coil_start_height),
                vertical_mid_point=(
                    self._tf_coil_start_radius, 0),
                thickness=self.outboard_tf_coil_radial_thickness,
                number_of_coils=self.number_of_tf_coils,
                distance=self.outboard_tf_coil_poloidal_thickness,
                name="tf_coil",
                rotation_angle=self.rotation_angle
            )
            comp = [self._tf_coil]
        return comp




import unittest

import paramak
import pytest


class TestInnerTfCoilsFlat(unittest.TestCase):

    def setUp(self):
        self.test_shape = paramak.InnerTfCoilsFlat(
            height=500,
            inner_radius=50,
            outer_radius=150,
            number_of_coils=6,
            gap_size=5
        )

        self.test_shape2 = paramak.InnerTfCoilsFlat(
            height=500,
            inner_radius=50,
            outer_radius=150,
            number_of_coils=6,
            gap_size=5,
            radius_type='straight'
        )

        # hexagon with 0 inner radius
        self.test_shape_3 = paramak.InnerTfCoilsFlat(
            height=10,
            inner_radius=0,
            outer_radius=20,
            number_of_coils=6,
            gap_size=0,
            radius_type='straight',
        )

    def test_default_parameters(self):
        """Checks that the default parameters of an InnerTfCoilsFlat are correct."""

        assert self.test_shape.rotation_angle == 360
        assert self.test_shape.azimuth_start_angle == 0
        assert self.test_shape.workplane == "XY"
        assert self.test_shape.rotation_axis == "Z"
        assert self.test_shape.radius_type == "corner"

    def test_points_calculation(self):
        """Checks that the points used to construct the InnerTfCoilsFlat
        component are calculated correctly from the parameters given."""

        assert self.test_shape.points == [
            (49.937460888595446, 2.5),
            (27.1320420790315, 41.99824154201773),
            (77.154447582418, 128.6358861991937),
            (149.97916521970643, 2.5)
        ]

    def test_processed_points_calculation(self):
        """Checks that the points used to construct the InnerTfCoilsFlat
        component are calculated correctly from the parameters given."""

        assert self.test_shape.processed_points == [
            (49.937460888595446, 2.5, 'straight'),
            (27.1320420790315, 41.99824154201773, 'straight'),
            (77.154447582418, 128.6358861991937, 'straight'),
            (149.97916521970643, 2.5, 'straight'),
            (49.937460888595446, 2.5, 'straight')
        ]

    def test_creation(self):
        """Creates an inner tf coil using the InnerTFCoilsFlat parametric
        component and checks that a cadquery solid is created."""

        assert self.test_shape.solid is not None
        assert self.test_shape.volume() > 1000

    def test_azimuth_offset(self):
        """Creates an inner tf coil using the InnerTfCoilsFlat parametric
        component and checks that the azimuthal start angle can be changed
        correctly."""

        assert self.test_shape.azimuth_placement_angle == [
            0, 60, 120, 180, 240, 300]
        self.test_shape.azimuth_start_angle = 20
        assert self.test_shape.azimuth_placement_angle == [
            20, 80, 140, 200, 260, 320]

    def test_attributes(self):
        """Checks that changing the attributes of InnerTfCoilsFlat affects the
        cadquery solid produced."""

        test_volume = self.test_shape.volume()

        self.test_shape.height = 1000
        assert test_volume == self.test_shape.volume() * 0.5
        self.test_shape.height = 500
        self.test_shape.inner_radius = 30
        assert test_volume < self.test_shape.volume()
        self.test_shape.inner_radius = 50
        self.test_shape.outer_radius = 170
        assert test_volume < self.test_shape.volume()

    def test_gap_size(self):
        """Checks that a ValueError is raised when a too large gap_size is
        used."""

        def test_incorrect_gap_size():
            self.test_shape.inner_radius = 20
            self.test_shape.outer_radius = 40
            self.test_shape.gap_size = 50
            self.test_shape.solid

        self.assertRaises(
            ValueError,
            test_incorrect_gap_size
        )

    def test_radius_type(self):
        """Checks that a ValueError is raised when radius_type is not a
        valid option."""

        def test_incorrect_radius_type():
            self.test_shape.radius_type = 'coucou'

        self.assertRaises(
            ValueError,
            test_incorrect_radius_type
        )

    def test_volume_changes_with_radius_type(self):
        """Checks the analytical volume of the hex shaped extrusion and
        that adding a hole reduces voulume."""

        hex_volume = self.test_shape_3.volume()
        hex_with_hole = self.test_shape_3
        hex_with_hole.inner_radius = 1

        assert hex_volume > hex_with_hole.volume()
        assert pytest.approx(hex_volume, abs=0.5) == 1385.6 * 10

    def test_volume_changes_with_radius_type_for_corners(self):
        """Checks the analytical volume of the hex shaped extrusion and
        that adding a hole reduces voulume when the radius is to the corners."""

        hex_shape = self.test_shape_3
        hex_shape.radius_type = 'corner'
        hex_volume = hex_shape.volume()

        assert pytest.approx(hex_volume, abs=0.5) == 1039.2 * 10

        hex_with_hole = self.test_shape_3
        hex_with_hole.radius_type = 'corner'
        hex_with_hole.inner_radius = 2

        assert hex_volume > hex_with_hole.volume()




from typing import Optional

from paramak import ExtrudeStraightShape


class PortCutterRectangular(ExtrudeStraightShape):
    """Creates an extruded shape with a rectangular section that is used to cut
    other components (eg. blanket, vessel,..) in order to create ports.

    Args:
        height: height (cm) of the port cutter.
        width: width (cm) of the port cutter.
        distance: extruded distance (cm) of the port cutter.
        center_point: Center point of the port cutter. Defaults to (0, 0).
        workplane: workplane in which the port cutters are created. Defaults
            to "ZY".
        rotation_axis: axis around which the port cutters are rotated and
            placed. Defaults to "Z".
        extrusion_start_offset (float, optional): the distance between 0 and
            the start of the extrusion. Defaults to 1..
        fillet_radius (float, optional): If not None, radius (cm) of fillets
            added to edges orthogonal to the Z direction. Defaults to None.
        name (str, optional): defaults to "rectangular_port_cutter".
    """

    def __init__(
        self,
        height: float,
        width: float,
        distance: float,
        center_point: Optional[tuple] = (0, 0),
        workplane: Optional[str] = "ZY",
        rotation_axis: Optional[str] = "Z",
        extrusion_start_offset: Optional[float] = 1.,
        fillet_radius: Optional[float] = None,
        name: Optional[str] = "rectangular_port_cutter",
        **kwargs
    ):

        super().__init__(
            workplane=workplane,
            rotation_axis=rotation_axis,
            extrusion_start_offset=extrusion_start_offset,
            extrude_both=False,
            name=name,
            distance=distance,
            **kwargs
        )

        self.height = height
        self.width = width
        self.center_point = center_point
        self.fillet_radius = fillet_radius

    def find_points(self):
        if self.workplane[0] < self.workplane[1]:
            parameter_1 = self.width
            parameter_2 = self.height
        else:
            parameter_1 = self.height
            parameter_2 = self.width

        points = [
            (-parameter_1 / 2, parameter_2 / 2),
            (parameter_1 / 2, parameter_2 / 2),
            (parameter_1 / 2, -parameter_2 / 2),
            (-parameter_1 / 2, -parameter_2 / 2),
        ]
        points = [(e[0] + self.center_point[0], e[1] +
                   self.center_point[1]) for e in points]

        self.points = points

    def add_fillet(self, solid):
        if "X" not in self.workplane:
            filleting_edge = "|X"
        if "Y" not in self.workplane:
            filleting_edge = "|Y"
        if "Z" not in self.workplane:
            filleting_edge = "|Z"

        if self.fillet_radius is not None and self.fillet_radius != 0:
            solid = solid.edges(filleting_edge).fillet(self.fillet_radius)

        return solid
